#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import ah_bootstrap
import pkg_resources
from setuptools import setup
from astropy_helpers.setup_helpers import register_commands, get_package_info
from astropy_helpers.version_helpers import generate_version_py

NAME = 'astropy_helpers'
VERSION = '2.0'
RELEASE = 'dev' not in VERSION

generate_version_py(NAME, VERSION, RELEASE, False, uses_git=not RELEASE)

# Use the updated version including the git rev count
from astropy_helpers.version import version as VERSION

cmdclass = register_commands(NAME, VERSION, RELEASE)
# This package actually doesn't use the Astropy test command
del cmdclass['test']

setup(
    name=pkg_resources.safe_name(NAME),  # astropy_helpers -> astropy-helpers
    version=VERSION,
    description='Utilities for building and installing Astropy, Astropy '
                'affiliated packages, and their respective documentation.',
    author='The Astropy Developers',
    author_email='astropy.team@gmail.com',
    license='BSD',
    url=' https://github.com/astropy/astropy-helpers',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Framework :: Setuptools Plugin',
        'Framework :: Sphinx :: Extension',
        'Framework :: Sphinx :: Theme',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Archiving :: Packaging'
    ],
    cmdclass=cmdclass,
    zip_safe=False,
    **get_package_info(exclude=['astropy_helpers.tests'])
)



cdef extern from "frame/src/cframe.h":
    ctypedef struct CFrame:
        pass

    double frame_hamiltonian(CFrame *fr, double t, double *qp, int n_dim) nogil
    void frame_gradient(CFrame *fr, double t, double *qp, int n_dim, double *dH) nogil
    void frame_hessian(CFrame *fr, double t, double *qp, int n_dim, double *d2H) nogil

cdef class CFrameWrapper:
    cdef CFrame cframe
    cdef double[::1] _params
    cpdef energy(self, double[:,::1] w, double[::1] t)
    cpdef gradient(self, double[:,::1] w, double[::1] t)
    cpdef hessian(self, double[:,::1] w, double[::1] t)




# coding: utf-8

""" Test utilities  """

from __future__ import division, print_function


# Third-party
import numpy as np

# Project
from ..util import peak_to_peak_period, estimate_dt_n_steps
from ...potential import Hamiltonian, NFWPotential
from ...units import galactic

def test_peak_to_peak_period():
    ntimes = 16384

    # trivial test
    for true_T in [1., 2., 4.123]:
        t = np.linspace(0,10.,ntimes)
        f = np.sin(2*np.pi/true_T * t)
        T = peak_to_peak_period(t, f)
        assert np.allclose(T, true_T, atol=1E-3)

    # modulated trivial test
    true_T = 2.
    t = np.linspace(0,10.,ntimes)
    f = np.sin(2*np.pi/true_T * t) + 0.1*np.cos(2*np.pi/(10*true_T) * t)
    T = peak_to_peak_period(t, f)
    assert np.allclose(T, true_T, atol=1E-3)

def test_estimate_dt_n_steps():
    nperiods = 128
    pot = NFWPotential.from_circular_velocity(v_c=1., r_s=10., units=galactic)
    w0 = [10.,0.,0.,0.,0.9,0.]
    dt,n_steps = estimate_dt_n_steps(w0, pot, n_periods=nperiods, n_steps_per_period=256,
                                     func=np.nanmin)

    orbit = Hamiltonian(pot).integrate_orbit(w0, dt=dt, n_steps=n_steps)
    T = orbit.estimate_period()
    assert int(round((orbit.t.max()/T).decompose().value)) == nperiods



from __future__ import absolute_import
from distutils.core import Extension
from astropy_helpers import setup_helpers

def get_extensions():
    exts = []

    # malloc
    mac_incl_path = "/usr/include/malloc"

    cfg = setup_helpers.DistutilsExtensionArgs()
    cfg['include_dirs'].append('numpy')
    cfg['include_dirs'].append(mac_incl_path)
    cfg['include_dirs'].append('gala/potential')
    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('gala/integrate/cyintegrators/leapfrog.pyx')
    cfg['sources'].append('gala/potential/potential/src/cpotential.c')
    exts.append(Extension('gala.integrate.cyintegrators.leapfrog', **cfg))

    cfg = setup_helpers.DistutilsExtensionArgs()
    cfg['include_dirs'].append('numpy')
    cfg['include_dirs'].append(mac_incl_path)
    cfg['include_dirs'].append('gala/potential')
    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('gala/potential/hamiltonian/src/chamiltonian.c')
    cfg['sources'].append('gala/potential/potential/src/cpotential.c')
    cfg['sources'].append('gala/integrate/cyintegrators/dop853.pyx')
    cfg['sources'].append('gala/integrate/cyintegrators/dopri/dop853.c')
    exts.append(Extension('gala.integrate.cyintegrators.dop853', **cfg))

    return exts

def get_package_data():
    return {'gala.integrate': ['*.pyx', '*.pxd', '*/*.pyx', '*/*.pxd',
                               'cyintegrators/*.c',
                               'cyintegrators/dopri/*.c', 'cyintegrators/dopri/*.h']}



# coding: utf-8

from __future__ import absolute_import, unicode_literals, division, print_function


# Third party
import astropy.units as u
import pytest
import numpy as np
from collections import OrderedDict

# This project
from ..core import PotentialBase, CompositePotential
from ..builtin import (KeplerPotential, HernquistPotential,
                       HenonHeilesPotential)

from ..ccompositepotential import CCompositePotential
from ....integrate import LeapfrogIntegrator, DOPRI853Integrator
from ....units import solarsystem, galactic

class CompositeHelper(object):

    def setup(self):
        self.units = solarsystem
        self.p1 = KeplerPotential(m=1.*u.Msun, units=self.units)
        self.p2 = HernquistPotential(m=0.5*u.Msun, c=0.1*u.au,
                                     units=self.units)

    def test_shit(self):
        potential = self.Cls(one=self.p1, two=self.p2)

        q = np.ascontiguousarray(np.array([[1.1,0,0]]).T)
        print("val", potential.value(q))

        q = np.ascontiguousarray(np.array([[1.1,0,0]]).T)
        print("grad", potential.gradient(q))

    def test_composite_create(self):
        potential = self.Cls()

        # Add a point mass with same unit system
        potential["one"] = KeplerPotential(units=self.units, m=1.)

        with pytest.raises(TypeError):
            potential["two"] = "derp"

        assert "one" in potential.parameters
        assert "m" in potential.parameters["one"]
        with pytest.raises(TypeError):
            potential.parameters["m"] = "derp"

    def test_plot_composite(self):
        # TODO: do image comparison or something to compare?

        potential = self.Cls()

        # Add a kepler potential and a harmonic oscillator
        potential["one"] = self.p1
        potential["two"] = self.p2

        grid = np.linspace(-5.,5)
        fig = potential.plot_contours(grid=(grid,0.,0.))
        # fig.savefig(os.path.join(plot_path, "composite_kepler_sho_1d.png"))

        fig = potential.plot_contours(grid=(grid,grid,0.))
        # fig.savefig(os.path.join(plot_path, "composite_kepler_sho_2d.png"))

    def test_integrate(self):
        potential = self.Cls()
        potential["one"] = self.p1
        potential["two"] = self.p2

        for Integrator in [DOPRI853Integrator, LeapfrogIntegrator]:
            w_cy = potential.integrate_orbit([1.,0,0, 0,2*np.pi,0], dt=0.01, n_steps=1000,
                                             Integrator=Integrator, cython_if_possible=True)
            w_py = potential.integrate_orbit([1.,0,0, 0,2*np.pi,0], dt=0.01, n_steps=1000,
                                             Integrator=Integrator, cython_if_possible=False)

            assert np.allclose(w_cy.xyz.value, w_py.xyz.value)
            assert np.allclose(w_cy.v_xyz.value, w_py.v_xyz.value)

# ------------------------------------------------------------------------

class TestComposite(CompositeHelper):
    Cls = CompositePotential

class TestCComposite(CompositeHelper):
    Cls = CCompositePotential

def test_failures():
    p = CCompositePotential()
    p['derp'] = KeplerPotential(m=1.*u.Msun, units=solarsystem)
    with pytest.raises(ValueError):
        p['jnsdfn'] = HenonHeilesPotential(units=solarsystem)

def test_lock():
    p = CompositePotential()
    p['derp'] = KeplerPotential(m=1.*u.Msun, units=solarsystem)
    p.lock = True
    with pytest.raises(ValueError): # try adding potential after lock
        p['herp'] = KeplerPotential(m=2.*u.Msun, units=solarsystem)

    p = CCompositePotential()
    p['derp'] = KeplerPotential(m=1.*u.Msun, units=solarsystem)
    p.lock = True
    with pytest.raises(ValueError): # try adding potential after lock
        p['herp'] = KeplerPotential(m=2.*u.Msun, units=solarsystem)

# -------------------------------------------------------

class MyPotential(PotentialBase):
    def __init__(self, m, x0, units=None):
        parameters = OrderedDict()
        parameters['m'] = m
        parameters['x0'] = np.array(x0)
        super(MyPotential, self).__init__(parameters=parameters,
                                          units=units)

    def _energy(self, x, t):
        m = self.parameters['m']
        x0 = self.parameters['x0']
        r = np.sqrt(np.sum((x-x0[None])**2, axis=1))
        return -m/r

    def _gradient(self, x, t):
        m = self.parameters['m']
        x0 = self.parameters['x0']
        r = np.sqrt(np.sum((x-x0[None])**2, axis=1))
        return m*(x-x0[None])/r**3

def test_add():
    """ Test adding potentials to get a composite """
    p1 = KeplerPotential(units=galactic, m=1*u.Msun)
    p2 = HernquistPotential(units=galactic,
                            m=1.E11, c=0.26)

    comp1 = CompositePotential()
    comp1['0'] = p1
    comp1['1'] = p2

    py_p1 = MyPotential(m=1., x0=[1.,0.,0.], units=galactic)
    py_p2 = MyPotential(m=4., x0=[-1.,0.,0.], units=galactic)

    # python + python
    new_p = py_p1 + py_p2
    assert isinstance(new_p, CompositePotential)
    assert not isinstance(new_p, CCompositePotential)
    assert len(new_p.keys()) == 2

    # python + python + python
    new_p = py_p1 + py_p2 + py_p2
    assert isinstance(new_p, CompositePotential)
    assert len(new_p.keys()) == 3

    # cython + cython
    new_p = p1 + p2
    assert isinstance(new_p, CCompositePotential)
    assert len(new_p.keys()) == 2

    # cython + python
    new_p = py_p1 + p2
    assert isinstance(new_p, CompositePotential)
    assert not isinstance(new_p, CCompositePotential)
    assert len(new_p.keys()) == 2

    # cython + cython + python
    new_p = p1 + p2 + py_p1
    assert isinstance(new_p, CompositePotential)
    assert not isinstance(new_p, CCompositePotential)
    assert len(new_p.keys()) == 3



# coding: utf-8

# Third-party
import astropy.units as u
import pytest

# Project
from .helpers import _TestBase
from .. import Hamiltonian
from ...potential.builtin import KeplerPotential
from ...frame.builtin import StaticFrame, ConstantRotatingFrame
from ....units import solarsystem, galactic, dimensionless

def test_init():
    p = KeplerPotential(m=1.)
    f = StaticFrame()
    H = Hamiltonian(potential=p, frame=f)

    p = KeplerPotential(m=1., units=solarsystem)
    f = StaticFrame(units=solarsystem)
    H = Hamiltonian(potential=p, frame=f)
    H = Hamiltonian(potential=p)

    p = KeplerPotential(m=1.)
    f = StaticFrame(galactic)
    with pytest.raises(ValueError):
        H = Hamiltonian(potential=p, frame=f)

    p = KeplerPotential(m=1., units=solarsystem)
    f = StaticFrame()
    with pytest.raises(ValueError):
        H = Hamiltonian(potential=p, frame=f)

    p = KeplerPotential(m=1., units=solarsystem)
    f = ConstantRotatingFrame(Omega=1./u.yr, units=solarsystem)
    with pytest.raises(ValueError):
        H = Hamiltonian(potential=p, frame=f)



# coding: utf-8
# cython: boundscheck=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

from __future__ import division, print_function


# Standard library
from collections import OrderedDict

# Third-party
from astropy.extern import six
from astropy.utils import InheritDocstrings
import numpy as np
cimport numpy as np
np.import_array()
import cython
cimport cython

from libc.stdio cimport printf

# Project
from .core import CompositePotential
from .cpotential import CPotentialBase
from .cpotential cimport CPotentialWrapper, CPotential

__all__ = ['CCompositePotential']

cdef class CCompositePotentialWrapper(CPotentialWrapper):

    def __init__(self, list potentials):
        cdef:
            CPotential cp
            CPotential tmp_cp
            int i
            CPotentialWrapper[::1] _cpotential_arr

        self._potentials = potentials
        _cpotential_arr = np.array(potentials)

        n_components = len(potentials)
        self._n_params = np.zeros(n_components, dtype=np.int32)
        for i in range(n_components):
            self._n_params[i] = _cpotential_arr[i]._n_params[0]

        cp.n_components = n_components
        cp.n_params = &(self._n_params[0])
        cp.n_dim = 0

        for i in range(n_components):
            tmp_cp = _cpotential_arr[i].cpotential
            cp.parameters[i] = &(_cpotential_arr[i]._params[0])
            cp.q0[i] = &(_cpotential_arr[i]._q0[0])
            cp.value[i] = tmp_cp.value[0]
            cp.density[i] = tmp_cp.density[0]
            cp.gradient[i] = tmp_cp.gradient[0]
            cp.hessian[i] = tmp_cp.hessian[0]

            if cp.n_dim == 0:
                cp.n_dim = tmp_cp.n_dim
            elif cp.n_dim != tmp_cp.n_dim:
                raise ValueError("Input potentials must have same number of coordinate dimensions")

        self.cpotential = cp

    def __reduce__(self):
        return (self.__class__, (list(self._potentials),))

class CCompositePotential(CPotentialBase, CompositePotential):

    def __init__(self, **potentials):
        CompositePotential.__init__(self, **potentials)

    def _reset_c_instance(self):
        self._potential_list = []
        for p in self.values():
            self._potential_list.append(p.c_instance)
        self.G = p.G
        self.c_instance = CCompositePotentialWrapper(self._potential_list)

    def __setitem__(self, *args, **kwargs):
        CompositePotential.__setitem__(self, *args, **kwargs)
        self._reset_c_instance()

    def __setstate__(self, state):
        # when rebuilding from a pickle, temporarily release lock to add components
        self.lock = False
        for name,potential in state:
            self[name] = potential
        self._reset_c_instance()
        self.lock = True

    def __reduce__(self):
        """ Properly package the object for pickling """
        return self.__class__, (), list(self.items())



# coding: utf-8

""" Base class for integrators. """

from __future__ import division, print_function


# Third-party
import numpy as np
from astropy.utils import InheritDocstrings
from astropy.extern import six

# This project
from ..units import UnitSystem, DimensionlessUnitSystem

__all__ = ["Integrator"]

@six.add_metaclass(InheritDocstrings)
class Integrator(object):

    def __init__(self, func, func_args=(), func_units=None):
        if not hasattr(func, '__call__'):
            raise ValueError("func must be a callable object, e.g., a function.")

        self.F = func
        self._func_args = func_args

        if func_units is not None and not isinstance(func_units, DimensionlessUnitSystem):
            func_units = UnitSystem(func_units)
        else:
            func_units = DimensionlessUnitSystem()
        self._func_units = func_units

    def _prepare_ws(self, w0, mmap, n_steps):
        """
        Decide how to make the return array. If mmap is False, this returns a
        full array of zeros, but with the correct shape as the output. If mmap
        is True, return a pointer to a memory-mapped array. The latter is
        particularly useful for integrating a large number of orbits or
        integrating a large number of time steps.
        """
        from ..dynamics import PhaseSpacePosition
        if not isinstance(w0, PhaseSpacePosition):
            w0 = PhaseSpacePosition.from_w(w0)

        arr_w0 = w0.w(self._func_units)

        self.ndim,self.norbits = arr_w0.shape
        self.ndim = self.ndim//2

        return_shape = (2*self.ndim,n_steps+1,self.norbits)
        if mmap is None:
            # create the return arrays
            ws = np.zeros(return_shape, dtype=float)

        else:
            if mmap.shape != return_shape:
                raise ValueError("Shape of memory-mapped array doesn't match expected shape of "
                                 "return array ({} vs {})".format(mmap.shape, return_shape))

            if not mmap.flags.writeable:
                raise TypeError("Memory-mapped array must be a writable mode, not '{}'"
                                .format(mmap.mode))

            ws = mmap

        return w0, arr_w0, ws

    def _handle_output(self, w0, t, w):
        """
        """
        if w.shape[-1] == 1:
            w = w[...,0]

        pos_unit = self._func_units['length']
        t_unit = self._func_units['time']
        vel_unit = pos_unit / t_unit

        from ..dynamics import Orbit
        orbit = Orbit(pos=w[:self.ndim]*pos_unit,
                      vel=w[self.ndim:]*vel_unit,
                      t=t*t_unit) # HACK: BADDDD

        return orbit

    def run(self):
        """
        Run the integrator starting from the specified phase-space position.
        The initial conditions ``w0`` should be a `~gala.dynamics.PhaseSpacePosition`
        instance.

        There are a few combinations of keyword arguments accepted for
        specifying the timestepping. For example, you can specify a fixed
        timestep (``dt``) and a number of steps (``n_steps``), or an array of
        times::

            dt, n_steps[, t1] : (numeric, int[, numeric])
                A fixed timestep dt and a number of steps to run for.
            dt, t1, t2 : (numeric, numeric, numeric)
                A fixed timestep dt, an initial time, and a final time.
            t : array_like
                An array of times to solve on.

        .. warning::

            Right now, this always returns a `~gala.dynamics.Orbit` instance.
            This is wrong and will change!

        .. todo::

            Allow specifying the return orbit class.

        Parameters
        ----------
        w0 : `~gala.dynamics.PhaseSpacePosition`
            Initial conditions.
        **time_spec
            Timestep information passed to
            `~gala.integrate.time_spec.parse_time_specification`.

        Returns
        -------
        orbit : `~gala.dynamics.Orbit`

        """
        pass



# coding: utf-8

from __future__ import division, print_function


# Standard library
import abc
from collections import OrderedDict
import warnings
import uuid

# Third-party
import numpy as np
from astropy.constants import G
import astropy.units as u
from astropy.utils import isiterable
import six

# Project
from ..common import CommonBase
from ...dynamics import PhaseSpacePosition
from ...util import ImmutableDict, atleast_2d
from ...units import DimensionlessUnitSystem

__all__ = ["PotentialBase", "CompositePotential"]

@six.add_metaclass(abc.ABCMeta)
class PotentialBase(CommonBase):
    """
    A baseclass for defining pure-Python gravitational potentials.

    Subclasses must define (at minimum) a method that evaluates
    the potential energy at a given position ``q``
    and time ``t``: ``_energy(q, t)``. For integration, the subclasses
    must also define a method to evaluate the gradient,
    ``_gradient(q,t)``. Optionally, they may also define methods
    to compute the density and hessian: ``_density()``, ``_hessian()``.
    """

    def __init__(self, parameters, origin=None, parameter_physical_types=None,
                 ndim=3, units=None):

        self.units = self._validate_units(units)

        if parameter_physical_types is None:
            parameter_physical_types = dict()
        self._ptypes = parameter_physical_types
        self.parameters = self._prepare_parameters(parameters, self._ptypes,
                                                   self.units)

        try:
            self.G = G.decompose(self.units).value
        except u.UnitConversionError:
            self.G = 1. # TODO: this is a HACK and could lead to user confusion

        self.ndim = ndim

        if origin is None:
            origin = np.zeros(self.ndim)
        self.origin = self._remove_units(origin)

    # ========================================================================
    # Abstract methods that must be implemented by subclasses
    #
    @abc.abstractmethod
    def _energy(self, q, t=0.):
        pass

    @abc.abstractmethod
    def _gradient(self, q, t=0.):
        pass

    def _density(self, q, t=0.):
        raise NotImplementedError("This Potential has no implemented density function.")

    def _hessian(self, q, t=0.):
        raise NotImplementedError("This Potential has no implemented Hessian.")

    # ========================================================================
    # Utility methods
    #
    def _remove_units(self, x):
        """
        Always returns an array. If a Quantity is passed in, it converts to the
        units associated with this object and returns the value.
        """
        if hasattr(x, 'unit'):
            x = x.decompose(self.units).value

        else:
            x = np.array(x)

        return x

    def _remove_units_prepare_shape(self, x):
        """
        This is similar to that implemented by `gala.potential.common.CommonBase`,
        but returns just the position if the input is a `PhaseSpacePosition`.
        """
        if hasattr(x, 'unit'):
            x = x.decompose(self.units).value

        elif isinstance(x, PhaseSpacePosition):
            x = x.cartesian.xyz.decompose(self.units).value

        x = atleast_2d(x, insert_axis=1).astype(np.float64)
        return x

    # ========================================================================
    # Core methods that use the above implemented functions
    #
    def energy(self, q, t=0.):
        """
        Compute the potential energy at the given position(s).

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            The position to compute the value of the potential. If the
            input position object has no units (i.e. is an `~numpy.ndarray`),
            it is assumed to be in the same unit system as the potential.

        Returns
        -------
        E : `~astropy.units.Quantity`
            The potential energy per unit mass or value of the potential.
        """
        q = self._remove_units_prepare_shape(q)
        orig_shape,q = self._get_c_valid_arr(q)
        t = self._validate_prepare_time(t, q)
        ret_unit = self.units['energy'] / self.units['mass']

        return self._energy(q, t=t).T.reshape(orig_shape[1:]) * ret_unit

    def gradient(self, q, t=0.):
        """
        Compute the gradient of the potential at the given position(s).

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            The position to compute the value of the potential. If the
            input position object has no units (i.e. is an `~numpy.ndarray`),
            it is assumed to be in the same unit system as the potential.

        Returns
        -------
        grad : `~astropy.units.Quantity`
            The gradient of the potential. Will have the same shape as
            the input position.
        """
        q = self._remove_units_prepare_shape(q)
        orig_shape,q = self._get_c_valid_arr(q)
        t = self._validate_prepare_time(t, q)
        ret_unit = self.units['length'] / self.units['time']**2
        return (self._gradient(q, t=t).T.reshape(orig_shape) * ret_unit).to(self.units['acceleration'])

    def density(self, q, t=0.):
        """
        Compute the density value at the given position(s).

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            The position to compute the value of the potential. If the
            input position object has no units (i.e. is an `~numpy.ndarray`),
            it is assumed to be in the same unit system as the potential.

        Returns
        -------
        dens : `~astropy.units.Quantity`
            The potential energy or value of the potential. If the input
            position has shape ``q.shape``, the output energy will have
            shape ``q.shape[1:]``.
        """
        q = self._remove_units_prepare_shape(q)
        orig_shape,q = self._get_c_valid_arr(q)
        t = self._validate_prepare_time(t, q)
        ret_unit = self.units['mass'] / self.units['length']**3
        return (self._density(q, t=t).T * ret_unit).to(self.units['mass density'])

    def hessian(self, q, t=0.):
        """
        Compute the Hessian of the potential at the given position(s).

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            The position to compute the value of the potential. If the
            input position object has no units (i.e. is an `~numpy.ndarray`),
            it is assumed to be in the same unit system as the potential.

        Returns
        -------
        hess : `~astropy.units.Quantity`
            The Hessian matrix of second derivatives of the potential. If the input
            position has shape ``q.shape``, the output energy will have shape
            ``(q.shape[0],q.shape[0]) + q.shape[1:]``. That is, an ``n_dim`` by
            ``n_dim`` array (matrix) for each position.
        """
        q = self._remove_units_prepare_shape(q)
        orig_shape,q = self._get_c_valid_arr(q)
        t = self._validate_prepare_time(t, q)
        ret_unit = 1 / self.units['time']**2
        hess = np.moveaxis(self._hessian(q, t=t), 0, -1)
        return hess.reshape((orig_shape[0], orig_shape[0]) + orig_shape[1:]) * ret_unit

    # ========================================================================
    # Convenience methods that make use the base methods
    #
    def acceleration(self, q, t=0.):
        """
        Compute the acceleration due to the potential at the given position(s).

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            Position to compute the acceleration at.

        Returns
        -------
        acc : `~astropy.units.Quantity`
            The acceleration. Will have the same shape as the input
            position array, ``q``.
        """
        return -self.gradient(q, t=t)

    def mass_enclosed(self, q, t=0.):
        """
        Estimate the mass enclosed within the given position by assuming the potential
        is spherical.

        Parameters
        ----------
        q : `~gala.dynamics.PhaseSpacePosition`, `~astropy.units.Quantity`, array_like
            Position(s) to estimate the enclossed mass.

        Returns
        -------
        menc : `~astropy.units.Quantity`
            Mass enclosed at the given position(s). If the input position
            has shape ``q.shape``, the output energy will have shape
            ``q.shape[1:]``.
        """
        q = self._remove_units_prepare_shape(q)
        orig_shape,q = self._get_c_valid_arr(q)
        t = self._validate_prepare_time(t, q)

        # small step-size in direction of q
        h = 1E-3 # MAGIC NUMBER

        # Radius
        r = np.sqrt(np.sum(q**2, axis=1))

        epsilon = h*q/r[:,np.newaxis]

        dPhi_dr_plus = self._energy(q + epsilon, t=t)
        dPhi_dr_minus = self._energy(q - epsilon, t=t)
        diff = (dPhi_dr_plus - dPhi_dr_minus)

        if isinstance(self.units, DimensionlessUnitSystem):
            Gee = 1.
        else:
            Gee = G.decompose(self.units).value

        Menc = np.abs(r*r * diff / Gee / (2.*h))
        Menc = Menc.reshape(orig_shape[1:])

        return Menc * self.units['mass']

    def circular_velocity(self, q, t=0.):
        """
        Estimate the circular velocity at the given position assuming the potential
        is spherical.

        Parameters
        ----------
        q : array_like, numeric
            Position(s) to estimate the circular velocity.

        Returns
        -------
        vcirc : `~astropy.units.Quantity`
            Circular velocity at the given position(s). If the input position
            has shape ``q.shape``, the output energy will have shape
            ``q.shape[1:]``.

        """
        q = self._remove_units_prepare_shape(q)

        # Radius
        r = np.sqrt(np.sum(q**2, axis=0)) * self.units['length']
        dPhi_dxyz = self.gradient(q, t=t)
        dPhi_dr = np.sum(dPhi_dxyz * q/r.value, axis=0)

        return self.units.decompose(np.sqrt(r * np.abs(dPhi_dr)))

    # ========================================================================
    # Python special methods
    #
    def __call__(self, q):
        return self.energy(q)

    def __repr__(self):
        pars = ""
        if not isinstance(self.parameters, OrderedDict):
            keys = sorted(self.parameters.keys()) # to ensure the order is always the same
        else:
            keys = self.parameters.keys()

        for k in keys:
            v = self.parameters[k].value
            par_fmt = "{}"
            post = ""

            if hasattr(v,'unit'):
                post = " {}".format(v.unit)
                v = v.value

            if isinstance(v, float):
                if v == 0:
                    par_fmt = "{:.0f}"
                elif np.log10(v) < -2 or np.log10(v) > 5:
                    par_fmt = "{:.2e}"
                else:
                    par_fmt = "{:.2f}"

            elif isinstance(v, int) and np.log10(v) > 5:
                par_fmt = "{:.2e}"

            pars += ("{}=" + par_fmt + post).format(k,v) + ", "

        if isinstance(self.units, DimensionlessUnitSystem):
            return "<{}: {} (dimensionless)>".format(self.__class__.__name__, pars.rstrip(", "))
        else:
            return "<{}: {} ({})>".format(self.__class__.__name__, pars.rstrip(", "), ",".join(map(str, self.units._core_units)))

    def __str__(self):
        return self.__class__.__name__

    def __add__(self, other):
        if not isinstance(other, PotentialBase):
            raise TypeError('Cannot add a {} to a {}'
                            .format(self.__class__.__name__,
                                    other.__class__.__name__))

        new_pot = CompositePotential()

        if isinstance(self, CompositePotential):
            for k,v in self.items():
                new_pot[k] = v

        else:
            k = str(uuid.uuid4())
            new_pot[k] = self

        if isinstance(other, CompositePotential):
            for k,v in self.items():
                if k in new_pot:
                    raise KeyError('Potential component "{}" already exists --'
                                   'duplicate key provided in potential '
                                   'addition')
                new_pot[k] = v

        else:
            k = str(uuid.uuid4())
            new_pot[k] = other

        return new_pot

    # ========================================================================
    # Convenience methods that do fancy things
    #
    def plot_contours(self, grid, filled=True, ax=None, labels=None,
                      subplots_kw=dict(), **kwargs):
        """
        Plot equipotentials contours. Computes the potential energy on a grid
        (specified by the array `grid`).

        .. warning:: Right now the grid input must be arrays and must already
            be in the unit system of the potential. Quantity support is coming...

        Parameters
        ----------
        grid : tuple
            Coordinate grids or slice value for each dimension. Should be a
            tuple of 1D arrays or numbers.
        filled : bool (optional)
            Use :func:`~matplotlib.pyplot.contourf` instead of
            :func:`~matplotlib.pyplot.contour`. Default is ``True``.
        ax : matplotlib.Axes (optional)
        labels : iterable (optional)
            List of axis labels.
        subplots_kw : dict
            kwargs passed to matplotlib's subplots() function if an axes object
            is not specified.
        kwargs : dict
            kwargs passed to either contourf() or plot().

        Returns
        -------
        fig : `~matplotlib.Figure`

        """

        import matplotlib.pyplot as plt
        from matplotlib import cm

        # figure out which elements are iterable, which are numeric
        _grids = []
        _slices = []
        for ii,g in enumerate(grid):
            if isiterable(g):
                _grids.append((ii,g))
            else:
                _slices.append((ii,g))

        # figure out the dimensionality
        ndim = len(_grids)

        # if ndim > 2, don't know how to handle this!
        if ndim > 2:
            raise ValueError("ndim > 2: you can only make contours on a 2D grid. For other "
                             "dimensions, you have to specify values to slice.")

        if ax is None:
            # default figsize
            fig, ax = plt.subplots(1, 1, **subplots_kw)
        else:
            fig = ax.figure

        if ndim == 1:
            # 1D curve
            x1 = _grids[0][1]
            r = np.zeros((len(_grids) + len(_slices), len(x1)))
            r[_grids[0][0]] = x1

            for ii,slc in _slices:
                r[ii] = slc

            Z = self.energy(r*self.units['length']).value
            ax.plot(x1, Z, **kwargs)

            if labels is not None:
                ax.set_xlabel(labels[0])
                ax.set_ylabel("potential")
        else:
            # 2D contours
            x1,x2 = np.meshgrid(_grids[0][1], _grids[1][1])
            shp = x1.shape
            x1,x2 = x1.ravel(), x2.ravel()

            r = np.zeros((len(_grids) + len(_slices), len(x1)))
            r[_grids[0][0]] = x1
            r[_grids[1][0]] = x2

            for ii,slc in _slices:
                r[ii] = slc

            Z = self.energy(r*self.units['length']).value

            # make default colormap not suck
            cmap = kwargs.pop('cmap', cm.Blues)
            if filled:
                cs = ax.contourf(x1.reshape(shp), x2.reshape(shp), Z.reshape(shp),
                                 cmap=cmap, **kwargs)
            else:
                cs = ax.contour(x1.reshape(shp), x2.reshape(shp), Z.reshape(shp),
                                cmap=cmap, **kwargs)

            if labels is not None:
                ax.set_xlabel(labels[0])
                ax.set_ylabel(labels[1])

        return fig

    def plot_density_contours(self, grid, filled=True, ax=None, labels=None,
                              subplots_kw=dict(), **kwargs):
        """
        Plot density contours. Computes the density on a grid
        (specified by the array `grid`).

        .. warning:: Right now the grid input must be arrays and must already be in
            the unit system of the potential. Quantity support is coming...

        Parameters
        ----------
        grid : tuple
            Coordinate grids or slice value for each dimension. Should be a
            tuple of 1D arrays or numbers.
        filled : bool (optional)
            Use :func:`~matplotlib.pyplot.contourf` instead of
            :func:`~matplotlib.pyplot.contour`. Default is ``True``.
        ax : matplotlib.Axes (optional)
        labels : iterable (optional)
            List of axis labels.
        subplots_kw : dict
            kwargs passed to matplotlib's subplots() function if an axes object
            is not specified.
        kwargs : dict
            kwargs passed to either contourf() or plot().

        Returns
        -------
        fig : `~matplotlib.Figure`

        """

        import matplotlib.pyplot as plt
        from matplotlib import cm

        # figure out which elements are iterable, which are numeric
        _grids = []
        _slices = []
        for ii,g in enumerate(grid):
            if isiterable(g):
                _grids.append((ii,g))
            else:
                _slices.append((ii,g))

        # figure out the dimensionality
        ndim = len(_grids)

        # if ndim > 2, don't know how to handle this!
        if ndim > 2:
            raise ValueError("ndim > 2: you can only make contours on a 2D grid. For other "
                             "dimensions, you have to specify values to slice.")

        if ax is None:
            # default figsize
            fig, ax = plt.subplots(1, 1, **subplots_kw)
        else:
            fig = ax.figure

        if ndim == 1:
            # 1D curve
            x1 = _grids[0][1]
            r = np.zeros((len(_grids) + len(_slices), len(x1)))
            r[_grids[0][0]] = x1

            for ii,slc in _slices:
                r[ii] = slc

            Z = self.density(r*self.units['length']).value
            ax.plot(x1, Z, **kwargs)

            if labels is not None:
                ax.set_xlabel(labels[0])
                ax.set_ylabel("potential")
        else:
            # 2D contours
            x1,x2 = np.meshgrid(_grids[0][1], _grids[1][1])
            shp = x1.shape
            x1,x2 = x1.ravel(), x2.ravel()

            r = np.zeros((len(_grids) + len(_slices), len(x1)))
            r[_grids[0][0]] = x1
            r[_grids[1][0]] = x2

            for ii,slc in _slices:
                r[ii] = slc

            Z = self.density(r*self.units['length']).value

            # make default colormap not suck
            cmap = kwargs.pop('cmap', cm.Blues)
            if filled:
                cs = ax.contourf(x1.reshape(shp), x2.reshape(shp), Z.reshape(shp),
                                 cmap=cmap, **kwargs)
            else:
                cs = ax.contour(x1.reshape(shp), x2.reshape(shp), Z.reshape(shp),
                                cmap=cmap, **kwargs)

            # cs.cmap.set_under('w')
            # cs.cmap.set_over('k')

            if labels is not None:
                ax.set_xlabel(labels[0])
                ax.set_ylabel(labels[1])

        return fig

    def integrate_orbit(self, *args, **kwargs):
        """

        .. warning:: This is now deprecated. Convenient orbit integration should
            happen using the `gala.potential.Hamiltonian` class. With a
            static reference frame, you just need to pass your potential
            in to the `~gala.potential.Hamiltonian` constructor.

        Integrate an orbit in the current potential using the integrator class
        provided. Uses same time specification as `Integrator.run()` -- see
        the documentation for `gala.integrate` for more information.

        Parameters
        ----------
        w0 : `~gala.dynamics.PhaseSpacePosition`, array_like
            Initial conditions.
        Integrator : `~gala.integrate.Integrator` (optional)
            Integrator class to use.
        Integrator_kwargs : dict (optional)
            Any extra keyword argumets to pass to the integrator class
            when initializing. Only works in non-Cython mode.
        cython_if_possible : bool (optional)
            If there is a Cython version of the integrator implemented,
            and the potential object has a C instance, using Cython
            will be *much* faster.
        **time_spec
            Specification of how long to integrate. See documentation
            for `~gala.integrate.parse_time_specification`.

        Returns
        -------
        orbit : `~gala.dynamics.Orbit`

        """

        warnings.warn("Use `Hamiltonian.integrate_orbit()` instead. If you are using a "
                      "static reference frame, you just need to pass your "
                      "potential object in to the Hamiltonian constructor to use, e.g., "
                      "orbit = Hamiltonian(potential).integrate_orbit(...).",
                      DeprecationWarning)

        from ..hamiltonian import Hamiltonian
        return Hamiltonian(self).integrate_orbit(*args, **kwargs)

    def total_energy(self, x, v):
        """
        Compute the total energy (per unit mass) of a point in phase-space
        in this potential. Assumes the last axis of the input position /
        velocity is the dimension axis, e.g., for 100 points in 3-space,
        the arrays should have shape (100,3).

        Parameters
        ----------
        x : array_like, numeric
            Position.
        v : array_like, numeric
            Velocity.
        """
        warnings.warn("Use the energy methods on Orbit objects instead. In a future "
                      "release this will be removed.", DeprecationWarning)

        v = atleast_2d(v, insert_axis=1)
        return self.energy(x) + 0.5*np.sum(v**2, axis=0)

    def save(self, f):
        """
        Save the potential to a text file. See :func:`~gala.potential.save`
        for more information.

        Parameters
        ----------
        f : str, file_like
            A filename or file-like object to write the input potential object to.

        """
        from .io import save
        save(self, f)

    # ========================================================================
    # Deprecated methods
    #
    def _value(self, q, t=0.):
        warnings.warn("Use `_energy()` instead.", DeprecationWarning)
        return self._energy(q, t=t)

    def value(self, *args, **kwargs):
        __doc__ = self.energy.__doc__
        warnings.warn("Use `energy()` instead.", DeprecationWarning)
        return self.energy(*args, **kwargs)

class CompositePotential(PotentialBase, OrderedDict):
    """
    A potential composed of several distinct components. For example,
    two point masses or a galactic disk and halo, each with their own
    potential model.

    A `CompositePotential` is created like a Python dictionary, e.g.::

        >>> p1 = SomePotential(func1) # doctest: +SKIP
        >>> p2 = SomePotential(func2) # doctest: +SKIP
        >>> cp = CompositePotential(component1=p1, component2=p2) # doctest: +SKIP

    This object actually acts like an `OrderedDict`, so if you want to
    preserve the order of the potential components, use::

        >>> cp = CompositePotential() # doctest: +SKIP
        >>> cp['component1'] = p1 # doctest: +SKIP
        >>> cp['component2'] = p2 # doctest: +SKIP

    You can also use any of the built-in `Potential` classes as
    components::

        >>> from gala.potential import HernquistPotential
        >>> cp = CompositePotential()
        >>> cp['spheroid'] = HernquistPotential(m=1E11, c=10., units=(u.kpc,u.Myr,u.Msun,u.radian))

    """
    def __init__(self, *args, **kwargs):
        self._units = None
        self.ndim = None

        if len(args) > 0 and isinstance(args[0], list):
            for k,v in args[0]:
                kwargs[k] = v
        else:
            for i,v in args:
                kwargs[str(i)] = v

        self.lock = False
        for v in kwargs.values():
            self._check_component(v)

        OrderedDict.__init__(self, **kwargs)

    def __setitem__(self, key, value):
        self._check_component(value)
        super(CompositePotential, self).__setitem__(key, value)

    def _check_component(self, p):
        if not isinstance(p, PotentialBase):
            raise TypeError("Potential components may only be Potential "
                            "objects, not {0}.".format(type(p)))

        if self.units is None:
            self._units = p.units
            self.ndim = p.ndim

        else:
            if sorted([str(x) for x in self.units]) != sorted([str(x) for x in p.units]):
                raise ValueError("Unit system of new potential component must match "
                                 "unit systems of other potential components.")

            if p.ndim != self.ndim:
                raise ValueError("All potential components must have the same number of "
                                 "phase-space dimensions ({} in this case)".format(self.ndim))

        if self.lock:
            raise ValueError("Potential object is locked - new components can only be"
                             " added to unlocked potentials.")

    @property
    def units(self):  # read-only
        return self._units

    @property
    def parameters(self):
        params = dict()
        for k,v in self.items():
            params[k] = v.parameters
        return ImmutableDict(**params)

    def _energy(self, q, t=0.):
        return np.sum([p._energy(q, t) for p in self.values()], axis=0)

    def _gradient(self, q, t=0.):
        return np.sum([p._gradient(q, t) for p in self.values()], axis=0)

    def _hessian(self, w, t=0.):
        return np.sum([p._hessian(w, t) for p in self.values()], axis=0)

    def _density(self, q, t=0.):
        return np.sum([p._density(q, t) for p in self.values()], axis=0)

    def __repr__(self):
        return "<CompositePotential {}>".format(",".join(self.keys()))



##################
# Toy Potentials #
##################
import numpy as np
from scipy.optimize import fmin_bfgs, leastsq, fmin_powell,fmin

# in units kpc, km/s and 10^11 M_solar
# Grav = 430091.5694
Grav = 430091.7270069976 # This was a bug in Sanders' code!

# Triaxial harmonic

def H_ho(x,omega):
    """ Simple harmonic oscillator Hamiltonian = 0.5 * omega**2 * x**2"""
    return 0.5*np.sum(x[3:]**2+(omega*x[:3])**2)


def angact_ho(x,omega):
    """ Calculate angle and action variable in sho potential with
    parameter omega """
    action = (x[3:]**2+(omega*x[:3])**2)/(2.*omega)
    angle = np.array([np.arctan(-x[3+i]/omega[i]/x[i]) if x[i]!=0. else -np.sign(x[3+i])*np.pi/2. for i in range(3)])
    for i in range(3):
        if(x[i]<0):
            angle[i]+=np.pi
    return np.concatenate((action,angle % (2.*np.pi)))


def deltaH_ho(omega,xsamples):
    if(np.any(omega<1e-5)):
        return np.nan
    H = 0.5*np.sum(xsamples.T[3:]**2,axis=0)+0.5*np.sum((omega[:3]*xsamples.T[:3].T)**2,axis=1)
    return H-np.mean(H)

def Jac_deltaH_ho(omega,xsamples):
    dHdparams = omega[:3]*xsamples.T[:3].T**2
    return dHdparams-np.mean(dHdparams,axis=0)

def findbestparams_ho(xsamples):
    """ Minimize sum of square differences of H_sho-<H_sho> for timesamples """
    return np.abs(leastsq(deltaH_ho,np.array([10.,10.,10.]), Dfun = Jac_deltaH_ho, args=(xsamples,))[0])[:3]


# Isochrone

def cart2spol(X):
    """ Performs coordinate transformation from cartesian
    to spherical polar coordinates with (r,phi,theta) having
    usual meanings. """
    x,y,z,vx,vy,vz=X
    r=np.sqrt(x*x+y*y+z*z)
    p=np.arctan2(y,x)
    t=np.arccos(z/r)
    vr=(vx*np.cos(p)+vy*np.sin(p))*np.sin(t)+np.cos(t)*vz
    vp=-vx*np.sin(p)+vy*np.cos(p)
    vt=(vx*np.cos(p)+vy*np.sin(p))*np.cos(t)-np.sin(t)*vz
    return np.array([r,p,t,vr,vp,vt])


def H_iso(x,params):
    """ Isochrone Hamiltonian = -GM/(b+sqrt(b**2+(r-r0)**2))"""
    #r = (np.sqrt(np.sum(x[:3]**2))-params[2])**2
    r = np.sum(x[:3]**2)
    return 0.5*np.sum(x[3:]**2)-Grav*params[0]/(params[1]+np.sqrt(params[1]**2+r))


def angact_iso(x,params):
    """ Calculate angle and action variable in isochrone potential with
    parameters params = (M,b) """
    GM = Grav*params[0]
    E = H_iso(x,params)
    r,p,t,vr,vphi,vt=cart2spol(x)
    st=np.sin(t)
    Lz=r*vphi*st
    L=np.sqrt(r*r*vt*vt+Lz*Lz/st/st)
    if(E>0.):  # Unbound
        return (np.nan,np.nan,np.nan,np.nan,np.nan,np.nan)
    Jr=GM/np.sqrt(-2*E)-0.5*(L+np.sqrt(L*L+4*GM*params[1]))
    action = np.array([Jr,Lz,L-abs(Lz)])

    c=GM/(-2*E)-params[1]
    e=np.sqrt(1-L*L*(1+params[1]/c)/GM/c)
    eta=np.arctan2(r*vr/np.sqrt(-2.*E),params[1]+c-np.sqrt(params[1]**2+r*r))
    OmR=np.power(-2*E,1.5)/GM
    Omp=0.5*OmR*(1+L/np.sqrt(L*L+4*GM*params[1]))
    thetar=eta-e*c*np.sin(eta)/(c+params[1])

    if(abs(vt)>1e-10):
        psi=np.arctan2(np.cos(t),-np.sin(t)*r*vt/L)
    else:
        psi=np.pi/2.
    a=np.sqrt((1+e)/(1-e))
    ap=np.sqrt((1+e+2*params[1]/c)/(1-e+2*params[1]/c))
    F = lambda x,y: np.pi/2.-np.arctan(np.tan(np.pi/2.-0.5*y)/x) if y>np.pi/2. \
        else -np.pi/2.+np.arctan(np.tan(np.pi/2.+0.5*y)/x) if y<-np.pi/2. \
        else np.arctan(x*np.tan(0.5*y))

    thetaz=psi+Omp*thetar/OmR-F(a,eta)-F(ap,eta)/np.sqrt(1+4*GM*params[1]/L/L)

    LR=Lz/L
    sinu = LR/np.sqrt(1.-LR**2)/np.tan(t)
    u = 0
    if(sinu>1.):
        u=np.pi/2.
    elif(sinu<-1.):
        u = -np.pi/2.
    else:
        u = np.arcsin(sinu)
    if(vt>0.):
        u=np.pi-u
    thetap=p-u+np.sign(Lz)*thetaz
    angle = np.array([thetar,thetap,thetaz])
    return np.concatenate((action,angle % (2.*np.pi)))


def deltaH_iso(params,p,r):
    deltaH = p-Grav*params[0]/(params[1]+np.sqrt(params[1]**2+r))
    if(params[0]<0. or params[1]<0. or np.any(deltaH>0.)):
        return np.nan
    return (deltaH-np.mean(deltaH))
    # return JR-np.mean(JR)


def Jac_deltaH_iso(params,p,r):
    H_o = -Grav/(params[1]+np.sqrt(params[1]**2+r))
    H_1 = Grav*params[0]*(1.+params[1]/np.sqrt(params[1]**2+r))/(params[1]+np.sqrt(params[1]**2+r))**2
    return np.array([(H_o-np.mean(H_o)),(H_1-np.mean(H_1))])


def findbestparams_iso(xsamples):
    """ Minimize sum of square differences of H_iso-<H_iso> for timesamples"""
    p = 0.5*np.sum(xsamples.T[3:]**2,axis=0)
    r = np.sum(xsamples.T[:3]**2,axis=0)
    return np.abs(leastsq(deltaH_iso,np.array([10.,10.]), Dfun = None , col_deriv=1,args=(p,r,))[0]) #Jac_deltaH_iso



# coding: utf-8
# cython: boundscheck=False
# cython: debug=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

""" Generate mock streams. """

from __future__ import division, print_function


# Standard library
import warnings

# Third-party
import numpy as np
cimport numpy as np
np.import_array()

from libc.math cimport sqrt
from cpython.exc cimport PyErr_CheckSignals

from ...integrate.cyintegrators.dop853 cimport dop853_step
from ...potential.potential.cpotential cimport CPotentialWrapper
from ...potential.frame.cframe cimport CFrameWrapper
from ...integrate.cyintegrators.leapfrog cimport c_init_velocity, c_leapfrog_step
from ._coord cimport (sat_rotation_matrix, to_sat_coords, from_sat_coords,
                      cyl_to_car, car_to_cyl)

from ...potential import Hamiltonian
from ...potential.frame import StaticFrame

__all__ = ['_mock_stream_dop853', '_mock_stream_leapfrog']

cdef extern from "frame/src/cframe.h":
    ctypedef struct CFrame:
        pass

cdef extern from "potential/src/cpotential.h":
    ctypedef struct CPotential:
        pass
    double c_d2_dr2(CPotential *p, double t, double *q, double *epsilon) nogil

cpdef _mock_stream_dop853(hamiltonian, double[::1] t, double[:,::1] prog_w,
                          int release_every,
                          _k_mean, _k_disp,
                          double G, _prog_mass, seed,
                          double atol=1E-10, double rtol=1E-10, int nmax=0):
    """
    _mock_stream_dop853(hamiltonian, t, prog_w, release_every, k_mean, k_disp, G, prog_mass, seed, atol, rtol, nmax)

    Generate a mock stellar stream using the Streakline method.

    Parameters
    ----------
    hamiltonian : `gala.potential.Hamiltonian`
        An instance of a ``Hamiltonian`` containing a gravitational potential.
    t : `numpy.ndarray`
        An array of times. Should have shape ``(ntimesteps,)``.
    prog_w : `numpy.ndarray`
        The 6D coordinates for the orbit of the progenitor system at all times.
        Should have shape ``(ntimesteps,6)``.
    release_every : int
        Release particles at the Lagrange points every X timesteps.
    k_mean : `numpy.ndarray`
        Array of mean ``k`` values (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    k_disp : `numpy.ndarray`
        Array of ``k`` value dispersions (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    G : numeric
        The value of the gravitational constant, G, in the unit system used.
    prog_mass : float or `numpy.ndarray`
        The mass of the progenitor or the mass at each time. Should be a scalar or have
        shape ``(ntimesteps,)``.
    seed : int (optional)
        A random number seed for initializing the particle positions.
    atol : numeric (optional)
        Passed to the integrator. Absolute tolerance parameter. Default is 1E-10.
    rtol : numeric (optional)
        Passed to the integrator. Relative tolerance parameter. Default is 1E-10.
    nmax : int (optional)
        Passed to the integrator.
    """

    if not isinstance(hamiltonian, Hamiltonian):
        raise TypeError("Input must be a Hamiltonian object, not {}".format(type(hamiltonian)))

    if not hamiltonian.c_enabled:
        raise TypeError("Input Hamiltonian object does not support C-level access.")

    cdef:
        int i, j, k # indexing
        int res # result from calling dop853
        int ntimes = t.shape[0] # number of times
        int nparticles # total number of test particles released

        unsigned ndim = prog_w.shape[1] # phase-space dimensionality
        unsigned ndim_2 = ndim / 2 # configuration-space dimensionality

        double dt0 = t[1] - t[0] # initial timestep

        double[::1] w_prime = np.zeros(6) # 6-position of stripped star
        double[::1] cyl = np.zeros(6) # 6-position in cylindrical coords
        double[::1] prog_w_prime = np.zeros(6) # 6-position of progenitor rotated
        double[::1] prog_cyl = np.zeros(6) # 6-position of progenitor in cylindrical coords

        # k-factors for parametrized model of Fardal et al. (2015)
        double[::1] ks = np.zeros(6)

        # used for figuring out how many orbits to integrate at any given release time
        unsigned this_ndim, this_norbits

        double Om # angular velocity squared
        double d, sigma_r # distance, dispersion in release positions
        double r_tide, menc, f # tidal radius, mass enclosed, f factor

        double[::1] eps = np.zeros(3) # used for 2nd derivative estimation
        double[:,::1] R = np.zeros((3,3)) # rotation matrix

        double[::1] prog_mass = np.ascontiguousarray(np.atleast_1d(_prog_mass))
        double[:,::1] k_mean = np.ascontiguousarray(np.atleast_2d(_k_mean))
        double[:,::1] k_disp = np.ascontiguousarray(np.atleast_2d(_k_disp))
        double[::1] mu_k
        double[::1] sigma_k

        # whoa, so many dots
        CPotential cp = (<CPotentialWrapper>(hamiltonian.potential.c_instance)).cpotential
        CFrame cf = (<CFrameWrapper>(hamiltonian.frame.c_instance)).cframe

    # figure out how many particles are going to be released into the "stream"
    if ntimes % release_every == 0:
        nparticles = 2 * (ntimes // release_every)
    else:
        nparticles = 2 * (ntimes // release_every + 1)

    # container for only current positions of all particles
    arr = np.empty((nparticles,ndim))
    cdef double[::1] w = arr.ravel()
    cdef double[:,::1] _w = arr

    # beginning times for each particle
    cdef double[::1] t1 = np.empty(nparticles)
    cdef double t_end = t[ntimes-1]

    # -------

    if seed is not None:
        np.random.seed(seed)

    # copy over initial conditions from progenitor orbit to each streakline star
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        for k in range(ndim):
            w[2*i*ndim + k] = prog_w[j,k]
            w[2*i*ndim + k + ndim] = prog_w[j,k]

        i += 1

    # now go back to each set of initial conditions and modify initial condition
    #   based on mock prescription
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        t1[2*i] = t[j]
        t1[2*i+1] = t[j]

        if prog_mass.shape[0] == 1:
            M = prog_mass[0]
        else:
            M = prog_mass[j]

        if k_mean.shape[0] == 1:
            mu_k = k_mean[0]
            sigma_k = k_disp[0]
        else:
            mu_k = k_mean[j]
            sigma_k = k_disp[j]

        # angular velocity
        d = sqrt(prog_w[j,0]*prog_w[j,0] +
                 prog_w[j,1]*prog_w[j,1] +
                 prog_w[j,2]*prog_w[j,2])
        Om = np.linalg.norm(np.cross(prog_w[j,:3], prog_w[j,3:]) / d**2)

        # gradient of potential in radial direction
        f = Om*Om - c_d2_dr2(&cp, t[j], &prog_w[j,0], &eps[0]) # TODO: do I need to use hamiltonian here?
        r_tide = (G*M / f)**(1/3.)

        # the rotation matrix to transform from satellite coords to normal
        sat_rotation_matrix(&prog_w[j,0], &R[0,0])
        to_sat_coords(&prog_w[j,0], &R[0,0], &prog_w_prime[0])
        car_to_cyl(&prog_w_prime[0], &prog_cyl[0])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        # eject stars at tidal radius with same angular velocity as progenitor
        cyl[0] = prog_cyl[0] + ks[0]*r_tide
        cyl[1] = prog_cyl[1] + ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] + ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        cyl[0] = prog_cyl[0] - ks[0]*r_tide
        cyl[1] = prog_cyl[1] - ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] - ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim + ndim])

        i += 1

    for i in range(nparticles):
        dop853_step(&cp, &cf, &_w[i,0], t1[i], t_end, dt0,
                    ndim, 1,
                    atol, rtol, nmax)

        PyErr_CheckSignals()

    return np.asarray(w).reshape(nparticles, ndim)

cpdef _mock_stream_leapfrog(hamiltonian, double[::1] t, double[:,::1] prog_w,
                            int release_every,
                            _k_mean, _k_disp,
                            double G, double _prog_mass, seed,
                            dt=None):
    """
    _mock_stream_leapfrog(cpotential, t, prog_w, release_every, k_mean, k_disp, G, prog_mass, seed)

    Generate a mock stellar stream using the Streakline method.

    Parameters
    ----------
    hamiltonian : `gala.potential.Hamiltonian`
        An instance of a ``Hamiltonian`` containing a gravitational potential.
    t : `numpy.ndarray`
        An array of times. Should have shape ``(ntimesteps,)``.
    prog_w : `numpy.ndarray`
        The 6D coordinates for the orbit of the progenitor system at all times.
        Should have shape ``(ntimesteps,6)``.
    release_every : int
        Release particles at the Lagrange points every X timesteps.
    k_mean : `numpy.ndarray`
        Array of mean ``k`` values (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    k_disp : `numpy.ndarray`
        Array of ``k`` value dispersions (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    G : numeric
        The value of the gravitational constant, G, in the unit system used.
    prog_mass : float or `numpy.ndarray`
        The mass of the progenitor or the mass at each time. Should be a scalar or have
        shape ``(ntimesteps,)``.
    seed : int (optional)
        A random number seed for initializing the particle positions.
    dt : float (optional)
        Timestep for integrating the orbits of each particle in the stream. Uses the
        same timestep as the progenitor orbit by default.
    """

    if not isinstance(hamiltonian, Hamiltonian):
        raise TypeError("Input must be a Hamiltonian object, not {}".format(type(hamiltonian)))

    if not hamiltonian.c_enabled:
        raise TypeError("Input Hamiltonian object does not support C-level access.")

    if not isinstance(hamiltonian.frame, StaticFrame):
        raise TypeError("Leapfrog integration is currently only supported "
                        "for StaticFrame, not {}."
                        .format(hamiltonian.frame.__class__.__name__))

    cdef:
        int i, j, k # indexing
        int ntimes = t.shape[0] # number of times
        int nparticles # total number of test particles released

        unsigned ndim = prog_w.shape[1] # phase-space dimensionality
        unsigned ndim_2 = ndim / 2 # configuration-space dimensionality

        double dt0 = t[1] - t[0] # timestep of progenitor orbit
        double _dt

        double[::1] w_prime = np.zeros(6) # 6-position of stripped star
        double[::1] cyl = np.zeros(6) # 6-position in cylindrical coords
        double[::1] prog_w_prime = np.zeros(6) # 6-position of progenitor rotated
        double[::1] prog_cyl = np.zeros(6) # 6-position of progenitor in cylindrical coords

        # k-factors for parametrized model of Fardal et al. (2015)
        double[::1] ks = np.zeros(6)

        # used for figuring out how many orbits to integrate at any given release time
        unsigned this_ndim, this_norbits

        double Om # angular velocity squared
        double d, sigma_r # distance, dispersion in release positions
        double r_tide, menc, f # tidal radius, mass enclosed, f factor

        double[::1] eps = np.zeros(3) # used for 2nd derivative estimation
        double[:,::1] R = np.zeros((3,3)) # rotation matrix

        double[::1] prog_mass = np.ascontiguousarray(np.atleast_1d(_prog_mass))
        double[:,::1] k_mean = np.ascontiguousarray(np.atleast_2d(_k_mean))
        double[:,::1] k_disp = np.ascontiguousarray(np.atleast_2d(_k_disp))
        double[::1] mu_k
        double[::1] sigma_k

        double[::1] v_jm1_2 = np.zeros(ndim_2) # temporary array to hold velocities at half step
        double[::1] grad = np.zeros(ndim_2) # temporary array to hold gradient

        CPotential cp = (<CPotentialWrapper>(hamiltonian.potential.c_instance)).cpotential

    if dt is None:
        _dt = dt0
    else:
        # TODO: implement support for non-default timestep for orbits
        # need to change the way integration is done below because of time array
        raise NotImplementedError()
        # _dt = float(dt)

    # figure out how many particles are going to be released into the "stream"
    if ntimes % release_every == 0:
        nparticles = 2 * (ntimes // release_every)
    else:
        nparticles = 2 * (ntimes // release_every + 1)

    # container for only current positions of all particles
    cdef double[::1] w = np.empty(nparticles*ndim)

    # beginning times for each particle
    cdef double[::1] t1 = np.empty(nparticles)
    cdef double t_end = t[ntimes-1]

    # -------

    if seed is not None:
        np.random.seed(seed)

    # copy over initial conditions from progenitor orbit to each streakline star
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        for k in range(ndim):
            w[2*i*ndim + k] = prog_w[j,k]
            w[2*i*ndim + k + ndim] = prog_w[j,k]

        i += 1

    # now go back to each set of initial conditions and modify initial condition
    #   based on mock prescription
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        t1[2*i] = t[j]
        t1[2*i+1] = t[j]

        if prog_mass.shape[0] == 1:
            M = prog_mass[0]
        else:
            M = prog_mass[j]

        if k_mean.shape[0] == 1:
            mu_k = k_mean[0]
            sigma_k = k_disp[0]
        else:
            mu_k = k_mean[j]
            sigma_k = k_disp[j]

        # angular velocity
        d = sqrt(prog_w[j,0]*prog_w[j,0] +
                 prog_w[j,1]*prog_w[j,1] +
                 prog_w[j,2]*prog_w[j,2])
        Om = np.linalg.norm(np.cross(prog_w[j,:3], prog_w[j,3:]) / d**2)

        # gradient of potential in radial direction
        f = Om*Om - c_d2_dr2(&cp, t[j], &prog_w[j,0], &eps[0])
        r_tide = (G*M / f)**(1/3.)

        # the rotation matrix to transform from satellite coords to normal
        sat_rotation_matrix(&prog_w[j,0], &R[0,0])
        to_sat_coords(&prog_w[j,0], &R[0,0], &prog_w_prime[0])
        car_to_cyl(&prog_w_prime[0], &prog_cyl[0])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        # eject stars at tidal radius with same angular velocity as progenitor
        cyl[0] = prog_cyl[0] + ks[0]*r_tide
        cyl[1] = prog_cyl[1] + ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] + ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        cyl[0] = prog_cyl[0] - ks[0]*r_tide
        cyl[1] = prog_cyl[1] - ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] - ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim + ndim])

        i += 1

    for i in range(nparticles):
        c_init_velocity(&cp, ndim_2, t1[0], _dt,
                        &w[i*ndim], &w[i*ndim+ndim_2], &v_jm1_2[0], &grad[0])

        ntimes = int((t_end - t1[i]) / _dt)
        for j in range(ntimes):
            for k in range(ndim_2):
                grad[k] = 0.

            c_leapfrog_step(&cp, ndim_2, t1[j], _dt,
                            &w[i*ndim], &w[i*ndim+ndim_2], &v_jm1_2[0], &grad[0])

        PyErr_CheckSignals()

    return np.asarray(w).reshape(nparticles, ndim)

cpdef _mock_stream_animate(snapshot_filename, hamiltonian,
                           double[::1] t, double[:,::1] prog_w,
                           int release_every,
                           _k_mean, _k_disp,
                           double G, _prog_mass, seed,
                           double atol=1E-10, double rtol=1E-10, int nmax=0,
                           int check_filesize=1):
    """
    _mock_stream_animate(filename, cpotential, t, prog_w, release_every, k_mean, k_disp, G, prog_mass, seed, atol, rtol, nmax)

    WARNING: only use this if you want to make an animation of a stream forming!

    Parameters
    ----------
    snapshot_filename : str
        The filename of the HDF5 snapshot file to write.
    cpotential : `gala.potential._CPotential`
        An instance of a ``_CPotential`` representing the gravitational potential.
    t : `numpy.ndarray`
        An array of times. Should have shape ``(ntimesteps,)``.
    prog_w : `numpy.ndarray`
        The 6D coordinates for the orbit of the progenitor system at all times.
        Should have shape ``(ntimesteps,6)``.
    release_every : int
        Release particles at the Lagrange points every X timesteps.
    k_mean : `numpy.ndarray`
        Array of mean ``k`` values (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    k_disp : `numpy.ndarray`
        Array of ``k`` value dispersions (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        ``(R,phi,z,vR,vphi,vz)``. If 1D, assumed constant in time. If 2D, time axis is axis 0.
    G : numeric
        The value of the gravitational constant, G, in the unit system used.
    prog_mass : float or `numpy.ndarray`
        The mass of the progenitor or the mass at each time. Should be a scalar or have
        shape ``(ntimesteps,)``.
    seed : int (optional)
        A random number seed for initializing the particle positions.
    atol : numeric (optional)
        Passed to the integrator. Absolute tolerance parameter. Default is 1E-10.
    rtol : numeric (optional)
        Passed to the integrator. Relative tolerance parameter. Default is 1E-10.
    nmax : int (optional)
        Passed to the integrator.
    check_filesize : bool (optional)
        Check the output filesize and warn the user if it is larger than
        10GB. Default = True.
    """

    import h5py

    if not isinstance(hamiltonian, Hamiltonian):
        raise TypeError("Input must be a Hamiltonian object, not {}".format(type(hamiltonian)))

    if not hamiltonian.c_enabled:
        raise TypeError("Input Hamiltonian object does not support C-level access.")

    cdef:
        int i, j, k # indexing
        int res # result from calling dop853
        int ntimes = t.shape[0] # number of times
        int nparticles # total number of test particles released

        unsigned ndim = prog_w.shape[1] # phase-space dimensionality
        unsigned ndim_2 = ndim / 2 # configuration-space dimensionality

        double dt0 = t[1] - t[0] # initial timestep

        double[::1] w_prime = np.zeros(6) # 6-position of stripped star
        double[::1] cyl = np.zeros(6) # 6-position in cylindrical coords
        double[::1] prog_w_prime = np.zeros(6) # 6-position of progenitor rotated
        double[::1] prog_cyl = np.zeros(6) # 6-position of progenitor in cylindrical coords

        # k-factors for parametrized model of Fardal et al. (2015)
        double[::1] ks = np.zeros(6)

        # used for figuring out how many orbits to integrate at any given release time
        unsigned this_ndim, this_norbits

        double Om # angular velocity squared
        double d, sigma_r # distance, dispersion in release positions
        double r_tide, menc, f # tidal radius, mass enclosed, f factor

        double[::1] eps = np.zeros(3) # used for 2nd derivative estimation
        double[:,::1] R = np.zeros((3,3)) # rotation matrix

        double[::1] prog_mass = np.ascontiguousarray(np.atleast_1d(_prog_mass))
        double[:,::1] k_mean = np.ascontiguousarray(np.atleast_2d(_k_mean))
        double[:,::1] k_disp = np.ascontiguousarray(np.atleast_2d(_k_disp))
        double[::1] mu_k
        double[::1] sigma_k

        double t_j

        # whoa, so many dots
        CPotential cp = (<CPotentialWrapper>(hamiltonian.potential.c_instance)).cpotential
        CFrame cf = (<CFrameWrapper>(hamiltonian.frame.c_instance)).cframe

    # figure out how many particles are going to be released into the "stream"
    if ntimes % release_every == 0:
        nparticles = 2 * (ntimes // release_every)
    else:
        nparticles = 2 * (ntimes // release_every + 1)

    # estimate size of output file and warn user if it's large
    if check_filesize:
        est_filesize_GB = nparticles * ntimes / 2 * 8 / 1024 / 1024 / 1024
        if est_filesize_GB >= 10.:
            warnings.warn("gala.dynamics.mockstream: Estimated output "
                          "filesize is >= 10 GB!")

    # container for only current positions of all particles
    cdef double[::1] w = np.empty(nparticles*ndim)
    cdef double[:,::1] one_particle_w = np.empty((ntimes,ndim))

    # beginning times for each particle
    cdef double[::1] t1 = np.empty(nparticles)
    cdef int[::1] all_ntimes = np.zeros(nparticles, dtype=np.int32)
    cdef double t_end = t[ntimes-1]

    # -------

    if seed is not None:
        np.random.seed(seed)

    # copy over initial conditions from progenitor orbit to each streakline star
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        for k in range(ndim):
            w[2*i*ndim + k] = prog_w[j,k]
            w[2*i*ndim + k + ndim] = prog_w[j,k]

        i += 1

    # now go back to each set of initial conditions and modify initial condition
    #   based on mock prescription
    i = 0
    for j in range(ntimes):
        if (j % release_every) != 0:
            continue

        t1[2*i] = t[j]
        t1[2*i+1] = t[j]
        all_ntimes[2*i] = ntimes - j
        all_ntimes[2*i+1] = ntimes - j

        if prog_mass.shape[0] == 1:
            M = prog_mass[0]
        else:
            M = prog_mass[j]

        if k_mean.shape[0] == 1:
            mu_k = k_mean[0]
            sigma_k = k_disp[0]
        else:
            mu_k = k_mean[j]
            sigma_k = k_disp[j]

        # angular velocity
        d = sqrt(prog_w[j,0]*prog_w[j,0] +
                 prog_w[j,1]*prog_w[j,1] +
                 prog_w[j,2]*prog_w[j,2])
        Om = np.linalg.norm(np.cross(prog_w[j,:3], prog_w[j,3:]) / d**2)

        # gradient of potential in radial direction
        f = Om*Om - c_d2_dr2(&cp, t[j], &prog_w[j,0], &eps[0])
        r_tide = (G*M / f)**(1/3.)

        # the rotation matrix to transform from satellite coords to normal
        sat_rotation_matrix(&prog_w[j,0], &R[0,0])
        to_sat_coords(&prog_w[j,0], &R[0,0], &prog_w_prime[0])
        car_to_cyl(&prog_w_prime[0], &prog_cyl[0])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        # eject stars at tidal radius with same angular velocity as progenitor
        cyl[0] = prog_cyl[0] + ks[0]*r_tide
        cyl[1] = prog_cyl[1] + ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] + ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim])

        for k in range(6):
            if sigma_k[k] > 0:
                ks[k] = np.random.normal(mu_k[k], sigma_k[k])
            else:
                ks[k] = mu_k[k]

        cyl[0] = prog_cyl[0] - ks[0]*r_tide
        cyl[1] = prog_cyl[1] - ks[1]*r_tide/prog_cyl[0]
        cyl[2] = ks[2]*r_tide/prog_cyl[0]
        cyl[3] = prog_cyl[3] + ks[3]*prog_cyl[3]
        cyl[4] = prog_cyl[4] - ks[0]*ks[4]*Om*r_tide
        cyl[5] = ks[5]*Om*r_tide
        cyl_to_car(&cyl[0], &w_prime[0])
        from_sat_coords(&w_prime[0], &R[0,0], &w[2*i*ndim + ndim])

        i += 1

    # create the output file
    with h5py.File(str(snapshot_filename), 'w') as h5f:
        h5f.create_dataset('pos', dtype='f8', shape=(ndim_2, ntimes, nparticles),
                           fillvalue=np.nan, compression='gzip', compression_opts=9)
        h5f.create_dataset('vel', dtype='f8', shape=(ndim_2, ntimes, nparticles),
                           fillvalue=np.nan, compression='gzip', compression_opts=9)
        h5f.create_dataset('t', data=np.array(t))

    for i in range(nparticles):
        t_j = t1[i]

        for k in range(ndim):
            one_particle_w[0,k] = w[i*ndim + k]

        for j in range(1,all_ntimes[i]+1,1):
            dop853_step(&cp, &cf, &w[i*ndim], t_j, t_j+dt0, dt0,
                        ndim, 1,
                        atol, rtol, nmax)

            PyErr_CheckSignals()

            for k in range(ndim):
                one_particle_w[j,k] = w[i*ndim + k]

            t_j = t_j+dt0

        with h5py.File(str(snapshot_filename), 'a') as h5f:
            j = ntimes - all_ntimes[i]
            h5f['pos'][:,j:,i] = np.array(one_particle_w[:all_ntimes[i], :ndim_2]).T
            h5f['vel'][:,j:,i] = np.array(one_particle_w[:all_ntimes[i], ndim_2:]).T




# coding: utf-8

""" Transform a proper motion to/from Galactic to/from ICRS coordinates """

from __future__ import division, print_function

# Standard library
import warnings

# Third-party
import numpy as np
import astropy.coordinates as coord

__all__ = ['transform_proper_motion', 'pm_gal_to_icrs', 'pm_icrs_to_gal']

def transform_proper_motion(coordinate, pm, new_frame):
    """
    Transform the proper motion vector ``pm`` at the sky coordinate
    ``coordinate`` to the desired coordinate frame ``new_frame``.

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`, :class:`~astropy.coordinates.BaseCoordinateFrame`
        An instance of an Astropy coordinate object.
    pm : :class:`~astropy.units.Quantity`, iterable
        Proper motion components (longitude, latitude) in the same frame
        as the input coordinates. Can either be an iterable of two
        :class:`~astropy.units.Quantity` objects or a single
        :class:`~astropy.units.Quantity` with shape (2,N).
        The proper motion in longitude is assumed to already include
        the cos(latitude) term.
    new_frame : :class:`~astropy.coordinates.BaseCoordinateFrame` subclass
        The desired coordinate frame of the proper motion.

    Returns
    -------
    new_pm : tuple
        A length-2 tuple containing the proper motion components
        (longitude, latitude) in the new frame. The longitude component is
        includes the cos(latitude) term.

    Examples
    --------

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> c = coord.SkyCoord(ra=196.5*u.degree, dec=-10.33*u.deg, distance=16.2*u.kpc)
        >>> pm = [-1.53, 3.5]*u.mas/u.yr
        >>> transform_proper_motion(c, pm, coord.Galactic) # doctest: +FLOAT_CMP
        <Quantity [-1.19944367, 3.62660101] mas / yr>

    """
    url = "http://docs.astropy.org/en/stable/coordinates/velocities.html"
    warnings.warn("This function is now deprecated. Use the velocity "
                  "transformation functionality in Astropy instead. For more "
                  "information, see: {0}".format(url), DeprecationWarning)

    if hasattr(coordinate, 'frame'):
        coordinate = coordinate.frame

    frame_cls = coordinate.__class__
    c = frame_cls(coordinate.data.with_differentials(
        coord.UnitSphericalCosLatDifferential(*pm)))
    new_c = c.transform_to(new_frame)
    diff = new_c.data.differentials['s']
    pm = np.vstack((diff.d_lon_coslat.value, diff.d_lat.value)) * diff.d_lat.unit

    return pm.reshape((2,) + c.shape)

# ----------------------------------------------------------------------------
# Deprecated:
#

def pm_gal_to_icrs(coordinate, pm):
    r"""
    Convert proper motion in Galactic coordinates (l,b) to
    ICRS coordinates (RA, Dec).

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`, :class:`~astropy.coordinates.BaseCoordinateFrame`
        An instance of an Astropy coordinate object. Can be in any
        frame that is transformable to ICRS coordinates.
    pm : :class:`~astropy.units.Quantity`, iterable
        Full description of proper motion in Galactic longitude and
        latitude. Can either be a tuple of two
        :class:`~astropy.units.Quantity` objects or a single
        :class:`~astropy.units.Quantity` with shape (2,N).
        The proper motion in longitude is assumed to be multipled by
        cosine of Galactic latitude, :math:`\mu_l\cos b`.

    Returns
    -------
    pm : :class:`~astropy.units.Quantity`
        An astropy :class:`~astropy.units.Quantity` object specifying the
        proper motion vector array in ICRS coordinates. Will have shape
        (2,N).

    Examples
    --------

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> c = coord.SkyCoord(ra=196.5*u.degree, dec=-10.33*u.deg, distance=16.2*u.kpc)
        >>> pm = [-1.53, 3.5]*u.mas/u.yr
        >>> pm_gal_to_icrs(c, pm) # doctest: +FLOAT_CMP
        <Quantity [-1.84741767, 3.34334366] mas / yr>

    """

    # Note: Deprecation warning gets emitted from calling
    #       transform_proper_motion below.
    g = coordinate.transform_to(coord.Galactic)
    return transform_proper_motion(g, pm, coord.ICRS)

def pm_icrs_to_gal(coordinate, pm):
    r"""
    Convert proper motion in ICRS coordinates (RA, Dec) to
    Galactic coordinates (l,b).

    Parameters
    ----------
    coordinate : :class:`~astropy.coordinates.SkyCoord`, :class:`~astropy.coordinates.BaseCoordinateFrame`
        An instance of an Astropy coordinate object. Can be in any
        frame that is transformable to ICRS coordinates.
    pm : :class:`~astropy.units.Quantity`, iterable
        Full description of proper motion in Right ascension (RA) and
        declination (Dec). Can either be a tuple of two
        :class:`~astropy.units.Quantity` objects or a single
        :class:`~astropy.units.Quantity` with shape (2,N).
        The proper motion in RA is assumed to be multipled by
        cosine of declination, :math:`\mu_\alpha\cos\delta`.

    Returns
    -------
    pm : :class:`~astropy.units.Quantity`
        An astropy :class:`~astropy.units.Quantity` object specifying the
        proper motion vector array in Galactic coordinates. Will have shape
        (2,N).

    Examples
    --------

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> c = coord.SkyCoord(ra=196.5*u.degree, dec=-10.33*u.deg, distance=16.2*u.kpc)
        >>> pm = [-1.84741767, 3.34334366]*u.mas/u.yr
        >>> pm_icrs_to_gal(c, pm) # doctest: +FLOAT_CMP
        <Quantity [-1.52999988, 3.49999973] mas / yr>

    """

    # Note: Deprecation warning gets emitted from calling
    #       transform_proper_motion below.
    i = coordinate.transform_to(coord.ICRS)
    return transform_proper_motion(i, pm, coord.Galactic)




# coding: utf-8
# cython: boundscheck=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

__all__ = ['CFrameBase']

# Third-party
import numpy as np
cimport numpy as np
np.import_array()

from .core import FrameBase
from ..potential.cpotential import _validate_pos_arr
from ..potential.cpotential cimport energyfunc, gradientfunc, hessianfunc
from ...dynamics import PhaseSpacePosition

cdef class CFrameWrapper:
    """ Wrapper class for C implementation of reference frames. """

    cpdef energy(self, double[:,::1] w, double[::1] t):
        """
        w should have shape (n, ndim).
        """
        cdef:
            int n, ndim, i
            CFrame cf = self.cframe
        n,ndim = _validate_pos_arr(w)

        cdef double [::1] pot = np.zeros(n)
        if len(t) == 1:
            for i in range(n):
                pot[i] = frame_hamiltonian(&cf, t[0], &w[i,0], ndim//2)
        else:
            for i in range(n):
                pot[i] = frame_hamiltonian(&cf, t[i], &w[i,0], ndim//2)


        return np.array(pot)

    cpdef gradient(self, double[:,::1] w, double[::1] t):
        """
        w should have shape (n, ndim).
        """
        cdef:
            int n, ndim, i
            CFrame cf = self.cframe
        n,ndim = _validate_pos_arr(w)

        cdef double[:,::1] dH = np.zeros((n, ndim))
        if len(t) == 1:
            for i in range(n):
                frame_gradient(&cf, t[0], &w[i,0], ndim//2, &dH[i,0])
        else:
            for i in range(n):
                frame_gradient(&cf, t[i], &w[i,0], ndim//2, &dH[i,0])


        return np.array(dH)

    cpdef hessian(self, double[:,::1] w, double[::1] t):
        """
        w should have shape (n, ndim).
        """
        cdef:
            int n, ndim, i
            CFrame cf = self.cframe
        n,ndim = _validate_pos_arr(w)

        cdef double[:,:,::1] d2H = np.zeros((n, ndim, ndim))
        if len(t) == 1:
            for i in range(n):
                frame_hessian(&cf, t[0], &w[i,0], ndim//2, &d2H[i,0,0])
        else:
            for i in range(n):
                frame_hessian(&cf, t[i], &w[i,0], ndim//2, &d2H[i,0,0])

        return np.array(d2H)

class CFrameBase(FrameBase):

    def __init__(self, Wrapper, parameters, units, parameter_physical_types=None, ndim=3):
        self.units = self._validate_units(units)

        if parameter_physical_types is None:
            parameter_physical_types = dict()
        self._ptypes = parameter_physical_types

        self.parameters = self._prepare_parameters(parameters, self._ptypes, self.units)
        self.c_parameters = np.ravel([v.value for v in self.parameters.values()])
        self.c_instance = Wrapper(*self.c_parameters)

        self.ndim = ndim

    def __str__(self):
        return self.__class__.__name__

    def _energy(self, q, t):
        return self.c_instance.energy(q, t=t)

    def _gradient(self, q, t):
        return self.c_instance.gradient(q, t=t)

    def _density(self, q, t):
        return self.c_instance.density(q, t=t)

    def _hessian(self, q, t):
        return self.c_instance.hessian(q, t=t)



# coding: utf-8

from __future__ import division, print_function

# Third-party
import astropy.units as u
from astropy.coordinates import Galactic, CartesianRepresentation
from astropy.tests.helper import quantity_allclose
import numpy as np
import pytest

# Project
from ..core import PhaseSpacePosition
from ...potential import Hamiltonian, HernquistPotential
from ...potential.frame import StaticFrame, ConstantRotatingFrame
from ...units import galactic, solarsystem
from ..representation_nd import (NDCartesianRepresentation,
                                 NDCartesianDifferential)

def test_init_repr():

    # Passing in x1, x2
    rep = NDCartesianRepresentation([1., 1.])
    assert rep.xyz.shape == (2,)

    # Passing in x1, x2
    rep = NDCartesianRepresentation(np.random.random(size=(2, 8)))
    assert rep.xyz.shape == (2,8)
    rep[:1]

    for n in range(1, 6+1):
        print('N: '+str(n))

        xs = np.random.uniform(size=(n, 16)) * u.one
        rep = NDCartesianRepresentation(xs)
        for i in range(1, n+1):
            assert hasattr(rep, 'x'+str(i))

        xs2 = rep.xyz
        assert quantity_allclose(xs, xs2)

        rep2 = rep[:8]

        assert rep.shape == (16,)
        assert rep2.shape == (8,)

def test_init_diff():

    # Passing in x1, x2
    rep = NDCartesianDifferential([1., 1.])
    assert rep.d_xyz.shape == (2,)
    with pytest.raises(TypeError):
        rep[:1]

    # Passing in x1, x2
    rep = NDCartesianDifferential(np.random.random(size=(2, 8)))
    assert rep.d_xyz.shape == (2,8)
    rep[:1]

    for n in range(1, 6+1):
        print('N: '+str(n))

        xs = np.random.uniform(size=(n, 16)) * u.one
        rep = NDCartesianDifferential(xs)
        for i in range(1, n+1):
            assert hasattr(rep, 'd_x'+str(i))

        xs2 = rep.d_xyz
        assert quantity_allclose(xs, xs2)

        rep2 = rep[:8]

        assert rep.shape == (16,)
        assert rep2.shape == (8,)



# coding: utf-8

""" Utilities for Potential classes """

from __future__ import division, print_function


# Third-party
import numpy as np

# Project
from .core import PotentialBase

__all__ = ['from_equation']

# def _classnamify(s):
#     s = [x.lower() for x in str(s).split()]
#     words = []
#     for word in s:
#         words.append(word.capitalize())
#     return "".join(words)

def from_equation(expr, vars, pars, name=None, hessian=False):
    r"""
    Create a potential class from an expression for the potential.

    .. note::

        This utility requires having `Sympy <http://www.sympy.org/>`_ installed.

    .. warning::

        These potentials are *not* pickle-able and cannot be written
        out to YAML files (using `~gala.potential.PotentialBase.save()`)

    Parameters
    ----------
    expr : :class:`sympy.core.expr.Expr`, str
        Either a ``Sympy`` expression, or a string that can be converted to
        a ``Sympy`` expression.
    vars : iterable
        An iterable of variable names in the expression.
    pars : iterable
        An iterable of parameter names in the expression.
    name : str (optional)
        The name of the potential class returned.
    hessian : bool (optional)
        Generate a function to compute the Hessian.

    Returns
    -------
    CustomPotential : `~gala.potential.PotentialBase`
        A potential class that represents the input equation. To instantiate the
        potential, use just like a normal class with parameters.

    Examples
    --------
    Here we'll create a potential class for the harmonic oscillator
    potential, :math:`\Phi(x) = \frac{1}{2}\,k\,x^2`::

        >>> Potential = from_equation("1/2*k*x**2", vars="x", pars="k",
        ...                           name='HarmonicOscillator')
        >>> p1 = Potential(k=1.)
        >>> p1
        <HarmonicOscillatorPotential: k=1.00 (dimensionless)>

    The potential class (and object) is a fully-fledged subclass of
    `~gala.potential.PotentialBase` and therefore has many useful methods.
    For example, to integrate an orbit::

        >>> orbit = p1.integrate_orbit([1.,0], dt=0.01, n_steps=1000)

    """
    try:
        import sympy
        from sympy.utilities.lambdify import lambdify
    except ImportError:
        raise ImportError("sympy is required to use 'from_equation()' "
                          "potential class creation.")

    # convert all input to Sympy objects
    expr = sympy.sympify(expr)
    vars = [sympy.sympify(v) for v in vars]
    var_names = [v.name for v in vars]
    pars = [sympy.sympify(p) for p in pars]
    par_names = [p.name for p in pars]
    ndim = len(vars)

    # Energy / value
    energyfunc = lambdify(vars + pars, expr, dummify=False, modules='numpy')

    # Gradient
    gradfuncs = []
    for var in vars:
        gradfuncs.append(lambdify(vars + pars, sympy.diff(expr,var), dummify=False, modules='numpy'))

    class CustomPotential(PotentialBase):

        def __init__(self, units=None, **kwargs):
            for par in par_names:
                if par not in kwargs:
                    raise ValueError("You must specify a value for "
                                     "parameter '{}'.".format(par))
            super(CustomPotential,self).__init__(units=units,
                                                 parameters=kwargs,
                                                 ndim=ndim)

        def _energy(self, w, t=0.):
            kw = self.parameters.copy()
            for k,v in kw.items():
                kw[k] = v.value

            for i,name in enumerate(var_names):
                kw[name] = w[:,i]

            return np.array(energyfunc(**kw))

        def _gradient(self, w, t=0.):
            kw = self.parameters.copy()
            for k,v in kw.items():
                kw[k] = v.value

            for i,name in enumerate(var_names):
                kw[name] = w[:,i]

            grad = np.vstack([f(**kw)[np.newaxis] for f in gradfuncs])
            return grad.T

    if name is not None:
        # name = _classnamify(name)
        if "potential" not in name.lower():
            name = name + "Potential"
        CustomPotential.__name__ = str(name)

    # Hessian
    if hessian:
        hessfuncs = []
        for var1 in vars:
            for var2 in vars:
                hessfuncs.append(lambdify(vars + pars, sympy.diff(expr,var1,var2),
                                          dummify=False, modules='numpy'))

        def _hessian(self, w, t):
            kw = self.parameters.copy()
            for k,v in kw.items():
                kw[k] = v.value

            for i,name in enumerate(var_names):
                kw[name] = w[:,i]

            # expand = [np.newaxis] * w[i].ndim

            # This ain't pretty, bub
            arrs = []
            for f in hessfuncs:
                hess_arr = np.array(f(**kw))
                if hess_arr.shape != w[:,i].shape:
                    hess_arr = np.tile(hess_arr, reps=w[:,i].shape)
                arrs.append(hess_arr)
            hess = np.vstack(arrs)

            return hess.reshape((ndim,ndim,len(w[:,i])))

        CustomPotential._hessian = _hessian

    CustomPotential.save = None
    return CustomPotential



# coding: utf-8

from __future__ import division, print_function

# Standard library
from collections import OrderedDict

# Third-party
import astropy.units as u
from astropy.utils import isiterable
import numpy as np

# Project
from ..dynamics import PhaseSpacePosition
from ..util import atleast_2d
from ..units import UnitSystem, DimensionlessUnitSystem

class CommonBase(object):

    def _validate_units(self, units):

        # make sure the units specified are a UnitSystem instance
        if units is not None and not isinstance(units, UnitSystem):
            units = UnitSystem(*units)

        elif units is None:
            units = DimensionlessUnitSystem()

        return units

    @classmethod
    def _prepare_parameters(cls, parameters, ptypes, units):
        pars = OrderedDict()
        for k,v in parameters.items():
            if hasattr(v, 'unit'):
                pars[k] = v.decompose(units)

            elif k in ptypes:
                # HACK TODO: remove when fix potentials that ask for scale velocity
                if ptypes[k] == 'speed':
                    pars[k] = v * units['length']/units['time']
                else:
                    pars[k] = v * units[ptypes[k]]

            else:
                pars[k] = v * u.one
        return pars

    def _remove_units_prepare_shape(self, x):
        if hasattr(x, 'unit'):
            x = x.decompose(self.units).value

        elif isinstance(x, PhaseSpacePosition):
            x = x.w(self.units)

        x = atleast_2d(x, insert_axis=1).astype(np.float64)
        return x

    def _get_c_valid_arr(self, x):
        """
        Warning! Interpretation of axes is different for C code.
        """
        orig_shape = x.shape
        x = np.ascontiguousarray(x.reshape(orig_shape[0], -1).T)
        return orig_shape, x

    def _validate_prepare_time(self, t, pos_c):
        """
        Make sure that t is a 1D array and compatible with the C position array.
        """
        if hasattr(t, 'unit'):
            t = t.decompose(self.units).value

        if not isiterable(t):
            t = np.atleast_1d(t)

        t = np.ascontiguousarray(t.ravel())

        if len(t) > 1:
            if len(t) != pos_c.shape[0]:
                raise ValueError("If passing in an array of times, it must have a shape "
                                 "compatible with the input position(s).")

        return t

    # For comparison operations
    def __eq__(self, other):
        # the funkiness in the below is in case there are array parameters:
        par_bool = [(k1==k2) and np.all(self.parameters[k1] == other.parameters[k2])
                    for k1,k2 in zip(self.parameters.keys(), other.parameters.keys())]
        return np.all(par_bool) and (str(self) == str(other)) and (self.units == other.units)

    def __ne__(self, other):
        return not self.__eq__(other)




# coding: utf-8

""" Test core dynamics.  """

from __future__ import division, print_function

# Standard library
import warnings

# Third-party
import astropy.units as u
from astropy.coordinates import (Galactic, CartesianRepresentation,
                                 SphericalRepresentation, CartesianDifferential,
                                 SphericalDifferential,
                                 SphericalCosLatDifferential)
from astropy.tests.helper import quantity_allclose
import numpy as np
import pytest

# Project
from ..core import PhaseSpacePosition
from ...potential import Hamiltonian, HernquistPotential
from ...potential.frame import StaticFrame, ConstantRotatingFrame
from ...units import galactic, solarsystem

def test_initialize():

    with pytest.raises(ValueError):
        x = np.random.random(size=(3,10))
        v = np.random.random(size=(3,8))
        PhaseSpacePosition(pos=x, vel=v)

    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    assert o.shape == (10,)

    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.random(size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    assert o.xyz.unit == u.kpc
    assert o.v_x.unit == u.km/u.s

    # Not 3D
    x = np.random.random(size=(2,10))
    v = np.random.random(size=(2,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    assert o.ndim == 2

    o = PhaseSpacePosition(pos=x, vel=v, frame=StaticFrame(galactic))
    assert o.ndim == 2
    assert o.frame is not None
    assert isinstance(o.frame, StaticFrame)

    x = np.random.random(size=(4,10))
    v = np.random.random(size=(4,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    assert o.ndim == 4

    # back to 3D
    pos = CartesianRepresentation(np.random.random(size=(3,10))*u.one)
    vel = CartesianDifferential(np.random.random(size=(3,10))*u.one)
    o = PhaseSpacePosition(pos=pos, vel=vel)
    assert hasattr(o, 'x')
    assert hasattr(o, 'y')
    assert hasattr(o, 'z')
    assert hasattr(o, 'v_x')
    assert hasattr(o, 'v_y')
    assert hasattr(o, 'v_z')

    # passing a representation with a differential attached
    pos = CartesianRepresentation(np.random.random(size=(3,10))*u.kpc)
    vel = CartesianDifferential(np.random.random(size=(3,10))*u.km/u.s)
    o = PhaseSpacePosition(pos.with_differentials({'s': vel}))
    assert hasattr(o, 'x')
    assert hasattr(o, 'y')
    assert hasattr(o, 'z')
    assert hasattr(o, 'v_x')
    assert hasattr(o, 'v_y')
    assert hasattr(o, 'v_z')

    o = o.represent_as(SphericalRepresentation)
    assert hasattr(o, 'distance')
    assert hasattr(o, 'lat')
    assert hasattr(o, 'lon')
    assert hasattr(o, 'radial_velocity')
    assert hasattr(o, 'pm_lon')
    assert hasattr(o, 'pm_lat')

    with pytest.raises(TypeError):
        o = PhaseSpacePosition(pos=x, vel=v, frame="blah blah blah")

    # check that old class raises deprecation warning
    from ..core import CartesianPhaseSpacePosition
    warnings.simplefilter('always')
    with pytest.warns(DeprecationWarning):
        o = CartesianPhaseSpacePosition(pos=x, vel=v)

def test_from_w():

    w = np.random.random(size=(6,10))
    o = PhaseSpacePosition.from_w(w, galactic)
    assert o.x.unit == u.kpc
    assert o.v_x.unit == u.kpc/u.Myr
    assert o.shape == (10,)

def test_slice():

    # simple
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    new_o = o[:5]
    assert new_o.shape == (5,)

    # 1d slice on 3d
    x = np.random.random(size=(3,10,8))
    v = np.random.random(size=(3,10,8))
    o = PhaseSpacePosition(pos=x, vel=v)
    new_o = o[:5]
    assert new_o.shape == (5,8)

    # 3d slice on 3d
    o = PhaseSpacePosition(pos=x, vel=v)
    new_o = o[:5,:4]
    assert new_o.shape == (5,4)

    # boolean array
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    ix = np.array([0,0,0,0,0,1,1,1,1,1]).astype(bool)
    new_o = o[ix]
    assert new_o.shape == (sum(ix),)

    # integer array
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    ix = np.array([0,3,5])
    new_o = o[ix]
    assert new_o.shape == (len(ix),)

# ------------------------------------------------------------------------
# Convert from Cartesian to other representations
# ------------------------------------------------------------------------
def test_represent_as():

    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    new_o = o.represent_as(SphericalRepresentation)
    o.spherical
    o.cylindrical
    o.cartesian

    assert new_o.pos.distance.unit == u.one
    assert new_o.vel.d_distance.unit == u.one

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    sph = o.represent_as(SphericalRepresentation)
    assert sph.pos.distance.unit == u.kpc

    sph2 = o.represent_as('spherical')
    for c in sph.pos.components:
        assert quantity_allclose(getattr(sph.pos, c),
                                 getattr(sph2.pos, c),
                                 rtol=1E-12)

    # doesn't work for 2D
    x = np.random.random(size=(2,10))
    v = np.random.random(size=(2,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    with pytest.raises(ValueError):
        o.represent_as(SphericalRepresentation)

def test_represent_as_expected_attributes():
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)

    new_o = o.spherical
    assert hasattr(new_o, 'distance')
    assert hasattr(new_o, 'lat')
    assert hasattr(new_o, 'lon')
    assert hasattr(new_o, 'radial_velocity')
    assert hasattr(new_o, 'pm_lat')
    assert hasattr(new_o, 'pm_lon')

    new_o = o.represent_as(SphericalRepresentation, SphericalCosLatDifferential)
    assert hasattr(new_o, 'distance')
    assert hasattr(new_o, 'lat')
    assert hasattr(new_o, 'lon')
    assert hasattr(new_o, 'radial_velocity')
    assert hasattr(new_o, 'pm_lat')
    assert hasattr(new_o, 'pm_lon_coslat')

    new_o = o.physicsspherical
    assert hasattr(new_o, 'r')
    assert hasattr(new_o, 'phi')
    assert hasattr(new_o, 'theta')
    assert hasattr(new_o, 'radial_velocity')
    assert hasattr(new_o, 'pm_theta')
    assert hasattr(new_o, 'pm_phi')

    new_o = o.cylindrical
    assert hasattr(new_o, 'rho')
    assert hasattr(new_o, 'phi')
    assert hasattr(new_o, 'z')
    assert hasattr(new_o, 'v_rho')
    assert hasattr(new_o, 'pm_phi')
    assert hasattr(new_o, 'v_z')

    new_o = new_o.cartesian
    assert hasattr(new_o, 'x')
    assert hasattr(new_o, 'y')
    assert hasattr(new_o, 'z')
    assert hasattr(new_o, 'xyz')
    assert hasattr(new_o, 'v_x')
    assert hasattr(new_o, 'v_y')
    assert hasattr(new_o, 'v_z')
    assert hasattr(new_o, 'v_xyz')

    # Check that this works with the NDCartesian classes too
    x = np.random.random(size=(2,10))*u.kpc
    v = np.random.normal(0.,100.,size=(2,10))*u.km/u.s
    new_o = PhaseSpacePosition(pos=x, vel=v)

    assert hasattr(new_o, 'x1')
    assert hasattr(new_o, 'x2')
    assert hasattr(new_o, 'xyz')
    assert hasattr(new_o, 'v_x1')
    assert hasattr(new_o, 'v_x2')
    assert hasattr(new_o, 'v_xyz')

def test_to_coord_frame():
    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)

    with pytest.raises(u.UnitConversionError):
        o.to_coord_frame(Galactic)

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    coo = o.to_coord_frame(Galactic)
    assert coo.name == 'galactic'

    warnings.simplefilter('always')
    with pytest.warns(DeprecationWarning):
        o.to_frame(Galactic)

    # doesn't work for 2D
    x = np.random.random(size=(2,10))*u.kpc
    v = np.random.normal(0.,100.,size=(2,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    with pytest.raises(ValueError):
        o.to_coord_frame(Galactic)

def test_w():
    # simple / unitless
    x = np.random.random(size=(3,10))
    v = np.random.random(size=(3,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    w = o.w()
    assert w.shape == (6,10)

    x = np.random.random(size=3)
    v = np.random.random(size=3)
    o = PhaseSpacePosition(pos=x, vel=v)
    w = o.w()
    assert w.shape == (6,1)

    # simple / unitless, 2D
    x = np.random.random(size=(2,10))
    v = np.random.random(size=(2,10))
    o = PhaseSpacePosition(pos=x, vel=v)
    w = o.w()
    assert w.shape == (4,10)

    x = np.random.random(size=2)
    v = np.random.random(size=2)
    o = PhaseSpacePosition(pos=x, vel=v)
    w = o.w()
    assert w.shape == (4,1)

    # simple / with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    with pytest.raises(ValueError):
        o.w()
    w = o.w(units=galactic)
    assert np.allclose(x.value, w[:3])
    assert np.allclose(v.value, (w[3:]*u.kpc/u.Myr).to(u.km/u.s).value)

    # simple / with units and potential
    p = HernquistPotential(units=galactic, m=1E11, c=0.25)
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    w = o.w(p.units)
    assert np.allclose(x.value, w[:3])
    assert np.allclose(v.value, (w[3:]*u.kpc/u.Myr).to(u.km/u.s).value)

    w = o.w(units=solarsystem)
    assert np.allclose(x.value, (w[:3]*u.au).to(u.kpc).value)
    assert np.allclose(v.value, (w[3:]*u.au/u.yr).to(u.km/u.s).value)

# ------------------------------------------------------------------------
# Computed dynamical quantities
# ------------------------------------------------------------------------
def test_energy():
    # with units
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    KE = o.kinetic_energy()
    assert KE.unit == (o.v_x.unit)**2
    assert KE.shape == o.shape

    # with units and potential
    p = HernquistPotential(units=galactic, m=1E11, c=0.25)
    H = Hamiltonian(p)
    x = np.random.random(size=(3,10))*u.kpc
    v = np.random.normal(0.,100.,size=(3,10))*u.km/u.s
    o = PhaseSpacePosition(pos=x, vel=v)
    PE = o.potential_energy(p)
    E = o.energy(H)

    warnings.simplefilter('always')
    with pytest.warns(DeprecationWarning):
        o.energy(p)

def test_angular_momentum():

    w = PhaseSpacePosition([1.,0.,0.], [0.,0.,1.])
    assert quantity_allclose(np.squeeze(w.angular_momentum()), [0.,-1,0]*u.one)

    w = PhaseSpacePosition([1.,0.,0.], [0.,1.,0.])
    assert quantity_allclose(np.squeeze(w.angular_momentum()), [0.,0, 1]*u.one)

    w = PhaseSpacePosition([0.,1.,0.],[0.,0.,1.])
    assert quantity_allclose(np.squeeze(w.angular_momentum()), [1., 0, 0]*u.one)

    w = PhaseSpacePosition([1.,0,0]*u.kpc, [0.,200.,0]*u.pc/u.Myr)
    assert quantity_allclose(np.squeeze(w.angular_momentum()), [0,0,0.2]*u.kpc**2/u.Myr)

    # multiple - known
    q = np.array([[1.,0.,0.],[1.,0.,0.],[0,1.,0.]]).T
    p = np.array([[0,0,1.],[0,1.,0.],[0,0,1]]).T
    L = PhaseSpacePosition(q, p).angular_momentum()
    true_L = np.array([[0., -1, 0],[0., 0, 1],[1., 0, 0]]).T * u.one
    assert L.shape == (3,3)
    assert quantity_allclose(L, true_L)

    # multiple - random
    q = np.random.uniform(size=(3,128))
    p = np.random.uniform(size=(3,128))
    L = PhaseSpacePosition(q, p).angular_momentum()
    assert L.shape == (3,128)

def test_frame_transform():
    static = StaticFrame(galactic)
    rotating = ConstantRotatingFrame(Omega=[0.53,1.241,0.9394]*u.rad/u.Myr,
                                     units=galactic)

    x = np.array([[10.,-0.2,0.3],[-0.232,8.1,0.1934]]).T * u.kpc
    v = np.array([[0.0034,0.2,0.0014],[0.0001,0.002532,-0.2]]).T * u.kpc/u.Myr

    # no frame specified at init
    psp = PhaseSpacePosition(pos=x, vel=v)
    with pytest.raises(ValueError):
        psp.to_frame(rotating)

    psp.to_frame(rotating, current_frame=static, t=0.4*u.Myr)

    # frame specified at init
    psp = PhaseSpacePosition(pos=x, vel=v, frame=static)
    psp.to_frame(rotating, t=0.4*u.Myr)

@pytest.mark.parametrize('obj', [
    PhaseSpacePosition([1,2,3.]*u.kpc, [1,2,3.]*u.km/u.s),
    PhaseSpacePosition([1,2,3.]*u.kpc, [1,2,3.]*u.km/u.s,
                       StaticFrame(galactic)),
    PhaseSpacePosition([1,2,3.]*u.kpc, [1,2,3.]*u.km/u.s,
                       ConstantRotatingFrame([1.,0,0]*u.rad/u.Myr, galactic)),
])
def test_io(tmpdir, obj):
    import h5py

    filename = str(tmpdir.join('thing.hdf5'))
    with h5py.File(filename, 'w') as f:
        obj.to_hdf5(f)

    obj2 = PhaseSpacePosition.from_hdf5(filename)
    assert quantity_allclose(obj.xyz, obj2.xyz)
    assert quantity_allclose(obj.v_xyz, obj2.v_xyz)
    assert obj.frame == obj2.frame



# coding: utf-8

""" Test dynamics plotting functions """

from __future__ import division, print_function

# Third-party
import numpy as np
import astropy.units as u

# Project
from ..core import PhaseSpacePosition
from ..orbit import Orbit
from ..plot import plot_projections

class TestPlotting(object):

    def setup(self):

        psps = []
        psps.append(PhaseSpacePosition(pos=np.random.random(size=3),
                                       vel=np.random.random(size=3)))
        psps.append(PhaseSpacePosition(pos=np.random.random(size=(3,16)),
                                       vel=np.random.random(size=(3,16))))
        psps.append(PhaseSpacePosition(pos=np.random.random(size=(3,16))*u.kpc,
                                       vel=np.random.random(size=(3,16))*u.km/u.s))

        orbits = []
        orbits.append(Orbit(pos=np.random.random(size=(3,16)),
                            vel=np.random.random(size=(3,16)),
                            t=np.linspace(0,1,16)))
        orbits.append(Orbit(pos=np.random.random(size=(3,16,2)),
                            vel=np.random.random(size=(3,16,2)),
                            t=np.linspace(0,1,16)))
        orbits.append(Orbit(pos=np.random.random(size=(3,16))*u.kpc,
                            vel=np.random.random(size=(3,16))*u.km/u.s,
                            t=np.linspace(0,1,16)*u.Myr))

        self.psps = psps
        self.orbits = orbits

    def test_plot_projections(self):
        import matplotlib.pyplot as plt

        # TODO: need major test improvements here
        # let the function create the figure
        for p in self.psps:
            _ = p.plot()

        for o in self.orbits:
            _ = o.plot()

        x = self.psps[0].xyz.value
        fig,axes = plt.subplots(1,2)
        fig = plot_projections(x[:2], autolim=True, axes=axes,
                               subplots_kwargs=dict(sharex=True),
                               labels=['x', 'y'],
                               plot_function=plt.plot,
                               marker='o', linestyle='--', color='r')


# def test_orbits(tmpdir):

#     # generate an "orbit"
#     n = 8
#     t = np.linspace(0, 100, 1000).reshape(1000,1)
#     x = np.cos(np.random.uniform(1.,8.,size=(1,n))*t).T[None]
#     y = np.sin(np.random.uniform(1.,8.,size=(1,n))*t).T[None]
#     z = np.cos(np.random.uniform(1.,8.,size=(1,n))*t).T[None]
#     vx = vy = vz = np.zeros_like(x)
#     w = np.rollaxis(np.vstack((x,y,z,vx,vy,vz)), -1, 1)

#     fig = plot_orbits(w, linestyle='none', marker='.', alpha=0.25)
#     fig.savefig(os.path.join(str(tmpdir), "all_orbits.png"))

#     fig = plot_orbits(w, ix=0, linestyle='none', marker='.', alpha=0.25)
#     fig.savefig(os.path.join(str(tmpdir), "one_orbit.png"))

#     fig = plot_orbits(w, ix=0, linestyle='none', marker='.', alpha=0.25,
#                       labels=("herp","derp","merp"))
#     fig.savefig(os.path.join(str(tmpdir), "one_orbit_labels.png"))

#     fig = plot_orbits(w, triangle=True, linestyle='-', marker=None)
#     fig.savefig(os.path.join(str(tmpdir), "all_orbits_triangle.png"))

#     fig = plot_orbits(w, ix=0, triangle=True, linestyle='-', marker=None)
#     fig.savefig(os.path.join(str(tmpdir), "one_orbit_triangle.png"))

#     fig = plot_orbits(w, ix=0, triangle=True, linestyle='-', marker=None,
#                       labels=("herp","derp","merp"))
#     fig.savefig(os.path.join(str(tmpdir), "one_orbit_triangle_labels.png"))

# def test_three_panel(tmpdir):

#     q = np.random.uniform(0.,10.,size=(1000,3))
#     q0 = np.array([5,5,5])

#     fig = three_panel(q)
#     fig.savefig(os.path.join(str(tmpdir), "three-panel-random.png"))

#     fig = three_panel(q, triangle=True)
#     fig.savefig(os.path.join(str(tmpdir), "three-panel-random_triangle.png"))

#     fig = three_panel(q, relative_to=q0, labels=[r'$\Omega_1$',r'$\Omega_2$',r'$\Omega_3$'])
#     fig.savefig(os.path.join(str(tmpdir), "three-panel-random-relative.png"))

#     fig = three_panel(q, relative_to=q0, triangle=True, labels=[r'$\Omega_1$',r'$\Omega_2$',r'$\Omega_3$'])
#     fig.savefig(os.path.join(str(tmpdir), "three-panel-random-relative_triangle.png"))

# def test_1d(tmpdir):

#     t = np.linspace(0,100.,1000)
#     q = np.cos(2*np.pi*t/10.)[None]

#     fig = plot_orbits(q, labels=(r"$\theta$",))
#     fig.savefig(os.path.join(str(tmpdir), "1d-orbit-labels.png"))

#     fig = plot_orbits(q, t=t, labels=(r"$\theta$",))
#     fig.savefig(os.path.join(str(tmpdir), "1d-orbit-labels-time.png"))

# def test_2d(tmpdir):

#     t = np.linspace(0,100.,1000)

#     q = np.zeros((2,len(t)))
#     q[0,:] = np.cos(2*np.pi*t/10.)
#     q[1,:] = np.sin(2*np.pi*t/5.5)

#     fig = plot_orbits(q, labels=(r"$\theta$",r"$\omega$"))
#     fig.savefig(os.path.join(str(tmpdir), "2d-orbit-labels.png"))



# coding: utf-8
"""
    Test the builtin CPotential classes
"""

from __future__ import absolute_import, unicode_literals, division, print_function


# Third-party
import numpy as np
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import pytest

# This project
from ..core import CompositePotential
from ..builtin import HenonHeilesPotential

def test_derp():
    p = HenonHeilesPotential()
    print(p.energy([0.77, 0.31]))
    print(p.density([0.77, 0.31]))



# coding: utf-8

from __future__ import division, print_function


# Standard library
import warnings
import os

# Third-party
import numpy as np

# Project
from .. import PhaseSpacePosition, Orbit
from ...potential import Hamiltonian, CPotentialBase
from ...integrate import DOPRI853Integrator, LeapfrogIntegrator
from ._mockstream import _mock_stream_dop853, _mock_stream_leapfrog, _mock_stream_animate

__all__ = ['mock_stream', 'streakline_stream', 'fardal_stream', 'dissolved_fardal_stream']

def mock_stream(hamiltonian, prog_orbit, prog_mass, k_mean, k_disp,
                release_every=1, Integrator=DOPRI853Integrator, Integrator_kwargs=dict(),
                snapshot_filename=None, seed=None):
    """
    Generate a mock stellar stream in the specified potential with a
    progenitor system that ends up at the specified position.

    Parameters
    ----------
    hamiltonian : `~gala.potential.Hamiltonian`
        The system Hamiltonian.
    prog_orbit : `~gala.dynamics.Orbit`
        The orbit of the progenitor system.
    prog_mass : numeric, array_like
        A single mass or an array of masses if the progenitor mass evolves
        with time.
    k_mean : `numpy.ndarray`
        Array of mean :math:`k` values (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        :math:`(R,\phi,z,v_R,v_\phi,v_z)`. If 1D, assumed constant in time. If 2D, time axis
        is axis 0.
    k_disp : `numpy.ndarray`
        Array of :math:`k` value dispersions (see Fardal et al. 2015). These are used to determine
        the exact prescription for generating the mock stream. The components are for:
        :math:`(R,\phi,z,v_R,v_\phi,v_z)`. If 1D, assumed constant in time. If 2D, time axis
        is axis 0.
    release_every : int (optional)
        Release particles at the Lagrange points every X timesteps.
    Integrator : `~gala.integrate.Integrator` (optional)
        Integrator to use.
    Integrator_kwargs : dict (optional)
        Any extra keyword argumets to pass to the integrator function.
    snapshot_filename : str (optional)
        Filename to save all incremental snapshots of particle positions and
        velocities. Warning: this can make very large files if you are not
        careful!
    seed : int (optional)
        A random number seed for initializing the particle positions.

    Returns
    -------
    stream : `~gala.dynamics.PhaseSpacePosition`

    """

    if isinstance(hamiltonian, CPotentialBase):
        warnings.warn("This function now expects a `Hamiltonian` instance instead of "
                      "a `PotentialBase` subclass instance. If you are using a "
                      "static reference frame, you just need to pass your "
                      "potential object in to the Hamiltonian constructor to use, e.g., "
                      "Hamiltonian(potential).", DeprecationWarning)

        hamiltonian = Hamiltonian(hamiltonian)

    # ------------------------------------------------------------------------
    # Some initial checks to short-circuit if input is bad
    if Integrator not in [LeapfrogIntegrator, DOPRI853Integrator]:
        raise ValueError("Only Leapfrog and dop853 integration is supported for"
                         " generating mock streams.")

    if not isinstance(hamiltonian, Hamiltonian) or not hamiltonian.c_enabled:
        raise TypeError("Input potential must be a CPotentialBase subclass.")

    if not isinstance(prog_orbit, Orbit):
        raise TypeError("Progenitor orbit must be an Orbit subclass.")

    if snapshot_filename is not None and Integrator != DOPRI853Integrator:
        raise ValueError("If saving snapshots, must use the DOP853Integrator.")

    k_mean = np.atleast_1d(k_mean)
    k_disp = np.atleast_1d(k_disp)

    if k_mean.ndim > 1:
        assert k_mean.shape[0] == prog_orbit.t.size
        assert k_disp.shape[0] == prog_orbit.t.size

    # ------------------------------------------------------------------------

    if prog_orbit.t[1] < prog_orbit.t[0]:
        raise ValueError("Progenitor orbit goes backwards in time. Streams can only "
                         "be generated on orbits that run forwards. Hint: you can "
                         "reverse the orbit with prog_orbit[::-1], but make sure the array"
                         "of k_mean values is ordered correctly.")

    c_w = np.squeeze(prog_orbit.w(hamiltonian.units)).T # transpose for Cython funcs
    prog_w = np.ascontiguousarray(c_w)
    prog_t = np.ascontiguousarray(prog_orbit.t.decompose(hamiltonian.units).value)
    if hasattr(prog_mass, 'unit'):
        prog_mass = prog_mass.decompose(hamiltonian.units).value

    if Integrator == LeapfrogIntegrator:
        stream_w = _mock_stream_leapfrog(hamiltonian, t=prog_t, prog_w=prog_w,
                                         release_every=release_every,
                                         _k_mean=k_mean, _k_disp=k_disp,
                                         G=hamiltonian.potential.G,
                                         _prog_mass=prog_mass, seed=seed,
                                         **Integrator_kwargs)

    elif Integrator == DOPRI853Integrator:
        if snapshot_filename is not None:
            if os.path.exists(snapshot_filename):
                raise IOError("Mockstream save file '{}' already exists.")

            import h5py

            _mock_stream_animate(snapshot_filename, hamiltonian, t=prog_t, prog_w=prog_w,
                                 release_every=release_every,
                                 _k_mean=k_mean, _k_disp=k_disp,
                                 G=hamiltonian.potential.G,
                                 _prog_mass=prog_mass, seed=seed,
                                 **Integrator_kwargs)

            with h5py.File(str(snapshot_filename), 'a') as h5f:
                h5f['pos'].attrs['unit'] = str(hamiltonian.units['length'])
                h5f['vel'].attrs['unit'] = str(hamiltonian.units['length']/hamiltonian.units['time'])
                h5f['t'].attrs['unit'] = str(hamiltonian.units['time'])

            return None

        else:
            stream_w = _mock_stream_dop853(hamiltonian, t=prog_t, prog_w=prog_w,
                                           release_every=release_every,
                                           _k_mean=k_mean, _k_disp=k_disp,
                                           G=hamiltonian.potential.G,
                                           _prog_mass=prog_mass, seed=seed,
                                           **Integrator_kwargs)

    else:
        raise RuntimeError("Should never get here...")

    return PhaseSpacePosition.from_w(w=stream_w.T, units=hamiltonian.units)

def streakline_stream(hamiltonian, prog_orbit, prog_mass, release_every=1,
                      Integrator=DOPRI853Integrator, Integrator_kwargs=dict(),
                      snapshot_filename=None, seed=None):
    """
    Generate a mock stellar stream in the specified potential with a
    progenitor system that ends up at the specified position.

    This uses the Streakline method from Kuepper et al. (2012).

    Parameters
    ----------
    hamiltonian : `~gala.potential.Hamiltonian`
        The system Hamiltonian.
    prog_orbit : `~gala.dynamics.Orbit`
            The orbit of the progenitor system.
    prog_mass : numeric, array_like
        A single mass or an array of masses if the progenitor mass evolves
        with time.
    release_every : int (optional)
        Release particles at the Lagrange points every X timesteps.
    Integrator : `~gala.integrate.Integrator` (optional)
        Integrator to use.
    Integrator_kwargs : dict (optional)
        Any extra keyword argumets to pass to the integrator function.
    snapshot_filename : str (optional)
        Filename to save all incremental snapshots of particle positions and
        velocities. Warning: this can make very large files if you are not
        careful!
    seed : int (optional)
        A random number seed for initializing the particle positions.

    Returns
    -------
    stream : `~gala.dynamics.PhaseSpacePosition`

    """
    k_mean = np.zeros(6)
    k_disp = np.zeros(6)

    k_mean[0] = 1. # R
    k_disp[0] = 0.

    k_mean[1] = 0. # phi
    k_disp[1] = 0.

    k_mean[2] = 0. # z
    k_disp[2] = 0.

    k_mean[3] = 0. # vR
    k_disp[3] = 0.

    k_mean[4] = 1. # vt
    k_disp[4] = 0.

    k_mean[5] = 0. # vz
    k_disp[5] = 0.

    return mock_stream(hamiltonian=hamiltonian, prog_orbit=prog_orbit, prog_mass=prog_mass,
                       k_mean=k_mean, k_disp=k_disp, release_every=release_every,
                       Integrator=Integrator, Integrator_kwargs=Integrator_kwargs,
                       snapshot_filename=snapshot_filename, seed=seed)

def fardal_stream(hamiltonian, prog_orbit, prog_mass, release_every=1,
                  Integrator=DOPRI853Integrator, Integrator_kwargs=dict(),
                  snapshot_filename=None, seed=None):
    """
    Generate a mock stellar stream in the specified potential with a
    progenitor system that ends up at the specified position.

    This uses the prescription from Fardal et al. (2015).

    Parameters
    ----------
    hamiltonian : `~gala.potential.Hamiltonian`
            The system Hamiltonian.
    prog_orbit : `~gala.dynamics.Orbit`
            The orbit of the progenitor system.
    prog_mass : numeric, array_like
        A single mass or an array of masses if the progenitor mass evolves
        with time.
    release_every : int (optional)
        Release particles at the Lagrange points every X timesteps.
    Integrator : `~gala.integrate.Integrator` (optional)
        Integrator to use.
    Integrator_kwargs : dict (optional)
        Any extra keyword argumets to pass to the integrator function.
    snapshot_filename : str (optional)
        Filename to save all incremental snapshots of particle positions and
        velocities. Warning: this can make very large files if you are not
        careful!
    seed : int (optional)
        A random number seed for initializing the particle positions.

    Returns
    -------
    stream : `~gala.dynamics.PhaseSpacePosition`

    """
    k_mean = np.zeros(6)
    k_disp = np.zeros(6)

    k_mean[0] = 2. # R
    k_disp[0] = 0.5

    k_mean[1] = 0. # phi
    k_disp[1] = 0.

    k_mean[2] = 0. # z
    k_disp[2] = 0.5

    k_mean[3] = 0. # vR
    k_disp[3] = 0.

    k_mean[4] = 0.3 # vt
    k_disp[4] = 0.5

    k_mean[5] = 0. # vz
    k_disp[5] = 0.5

    return mock_stream(hamiltonian=hamiltonian, prog_orbit=prog_orbit, prog_mass=prog_mass,
                       k_mean=k_mean, k_disp=k_disp, release_every=release_every,
                       Integrator=Integrator, Integrator_kwargs=Integrator_kwargs,
                       snapshot_filename=snapshot_filename, seed=seed)

def dissolved_fardal_stream(hamiltonian, prog_orbit, prog_mass, t_disrupt, release_every=1,
                            Integrator=DOPRI853Integrator, Integrator_kwargs=dict(),
                            snapshot_filename=None, seed=None):
    """
    Generate a mock stellar stream in the specified potential with a
    progenitor system that ends up at the specified position.

    This uses the prescription from Fardal et al. (2015), but at a specified
    time the progenitor completely dissolves and the radial offset of the
    tidal radius is reduced to 0 at fixed dispersion.

    Parameters
    ----------
    hamiltonian : `~gala.potential.Hamiltonian`
            The system Hamiltonian.
    prog_orbit : `~gala.dynamics.Orbit`
            The orbit of the progenitor system.
    prog_mass : numeric, array_like
        A single mass or an array of masses if the progenitor mass evolves
        with time.
    t_disrupt : numeric
        The time that the progenitor completely disrupts.
    release_every : int (optional)
        Release particles at the Lagrange points every X timesteps.
    Integrator : `~gala.integrate.Integrator` (optional)
        Integrator to use.
    Integrator_kwargs : dict (optional)
        Any extra keyword argumets to pass to the integrator function.
    snapshot_filename : str (optional)
        Filename to save all incremental snapshots of particle positions and
        velocities. Warning: this can make very large files if you are not
        careful!
    seed : int (optional)
        A random number seed for initializing the particle positions.

    Returns
    -------
    stream : `~gala.dynamics.PhaseSpacePosition`

    """

    try:
        # the time index closest to when the disruption happens
        t = prog_orbit.t
    except AttributeError:
        raise TypeError("Input progenitor orbit must be an Orbit subclass instance.")

    disrupt_ix = np.abs(t - t_disrupt).argmin()

    k_mean = np.zeros((t.size,6))
    k_disp = np.zeros((t.size,6))

    k_mean[:,0] = 2. # R
    k_mean[disrupt_ix:,0] = 0.
    k_disp[:,0] = 0.5

    k_mean[:,1] = 0. # phi
    k_disp[:,1] = 0.

    k_mean[:,2] = 0. # z
    k_disp[:,2] = 0.5

    k_mean[:,3] = 0. # vR
    k_disp[:,3] = 0.

    k_mean[:,4] = 0.3 # vt
    k_disp[:,4] = 0.5

    k_mean[:,5] = 0. # vz
    k_disp[:,5] = 0.5

    return mock_stream(hamiltonian=hamiltonian, prog_orbit=prog_orbit, prog_mass=prog_mass,
                       k_mean=k_mean, k_disp=k_disp, release_every=release_every,
                       Integrator=Integrator, Integrator_kwargs=Integrator_kwargs,
                       snapshot_filename=snapshot_filename, seed=seed)



# coding: utf-8

""" Test action-angle stuff """

from __future__ import division, print_function


# Standard library
import logging

# Third-party
import numpy as np
from astropy import log as logger
from astropy.tests.helper import quantity_allclose
from scipy.linalg import solve
import pytest

# Project
from ...integrate import DOPRI853Integrator
from ...potential import (IsochronePotential, HarmonicOscillatorPotential,
                          LeeSutoTriaxialNFWPotential)
from ...units import galactic
from ..actionangle import *
from ..core import *
from ..plot import *
from .helpers import *

logger.setLevel(logging.DEBUG)

def test_generate_n_vectors():
    # test against Sanders' method
    nvecs = generate_n_vectors(N_max=6, dx=2, dy=2, dz=2)
    nvecs_sanders = sanders_nvecs(N_max=6, dx=2, dy=2, dz=2)
    assert np.all(nvecs == nvecs_sanders)

    nvecs = generate_n_vectors(N_max=6, dx=1, dy=1, dz=1)
    nvecs_sanders = sanders_nvecs(N_max=6, dx=1, dy=1, dz=1)
    assert np.all(nvecs == nvecs_sanders)

def test_fit_isochrone():
    # integrate orbit in Isochrone potential, then try to recover it
    true_m = 2.81E11
    true_b = 11.
    potential = IsochronePotential(m=true_m, b=true_b, units=galactic)
    orbit = potential.integrate_orbit([15.,0,0,0,0.2,0], dt=2., n_steps=10000)

    fit_potential = fit_isochrone(orbit)
    m,b = fit_potential.parameters['m'].value, fit_potential.parameters['b'].value
    assert np.allclose(m, true_m, rtol=1E-2)
    assert np.allclose(b, true_b, rtol=1E-2)

def test_fit_harmonic_oscillator():
    # integrate orbit in harmonic oscillator potential, then try to recover it
    true_omegas = np.array([0.011, 0.032, 0.045])
    potential = HarmonicOscillatorPotential(omega=true_omegas, units=galactic)

    orbit = potential.integrate_orbit([15.,1,2,0,0,0], dt=2., n_steps=10000)

    fit_potential = fit_harmonic_oscillator(orbit)
    omegas = fit_potential.parameters['omega'].value
    assert np.allclose(omegas, true_omegas, rtol=1E-2)

# TODO: check on this after fixing HarmonicOscillatorPotential
def test_fit_toy_potential():
    # integrate orbit in both toy potentials, make sure correct one is chosen
    true_m = 2.81E11
    true_b = 11.
    true_potential = IsochronePotential(m=true_m, b=true_b, units=galactic)
    orbit = true_potential.integrate_orbit([15.,0,0,0,0.2,0], dt=2., n_steps=10000)

    potential = fit_toy_potential(orbit)
    for k,v in true_potential.parameters.items():
        assert quantity_allclose(v, potential.parameters[k], rtol=1E-2)

    # -----------------------------------------------------------------
    true_omegas = np.array([0.011, 0.032, 0.045])
    true_potential = HarmonicOscillatorPotential(omega=true_omegas, units=galactic)
    orbit = true_potential.integrate_orbit([15.,1,2,0,0,0], dt=2., n_steps=10000)

    potential = fit_toy_potential(orbit)

    assert quantity_allclose(potential.parameters['omega'],
                             true_potential.parameters['omega'], rtol=1E-2)

def test_check_angle_sampling():

    # frequencies
    omegas = np.array([0.21, 0.3421, 0.4968])

    # integer vectors
    nvecs = generate_n_vectors(N_max=6)

    # loop over times with known failures:
    #   - first one fails needing longer integration time
    #   - second one fails needing finer sampling
    for i,t in enumerate([np.linspace(0,50,500), np.linspace(0,8000,8000)]):
        # periods = 2*np.pi/omegas
        # print("Periods:", periods)
        # print("N periods:", t.max() / periods)

        angles = t[np.newaxis] * omegas[:,np.newaxis]
        checks,failures = check_angle_sampling(nvecs, angles)

        assert np.all(failures == i)

class ActionsBase(object):

    @pytest.fixture(autouse=True)
    def set_tmpdir(self, tmpdir):
        self.tmpdir = tmpdir

    def test_classify(self):
        # my classify
        orb_type = self.orbit.circulation()

        # compare to Sanders'
        for j in range(self.N):
            sdrs = genfunc_3d.assess_angmom(self.w[...,j].T)
            logger.debug("APW: {}, Sanders: {}".format(orb_type[:,j], sdrs))
            assert np.all(orb_type[:,j] == sdrs)

    def test_actions(self):
        # t = self.t[::10]
        t = self.t

        N_max = 6
        for n in range(self.N):
            print("\n\n")
            print("======================= Orbit {} =======================".format(n))
            # w = self.w[:,::10,n]
            w = self.w[...,n]
            orb = self.orbit[:,n]
            circ = orb.circulation()

            # get values from Sanders' code
            print("Computing actions from genfunc...")
            s_actions,s_angles,s_freqs,toy_potential = sanders_act_ang_freq(t, w, circ, N_max=N_max)

            print("Computing actions with gala...")
            ret = find_actions(orb, N_max=N_max, toy_potential=toy_potential)
            actions = ret['actions']
            angles = ret['angles']
            freqs = ret['freqs']

            print("Action ratio: {}".format(actions / s_actions))
            print("Angle ratio: {}".format(angles / s_angles))
            print("Freq ratio: {}".format(freqs / s_freqs))

            assert np.allclose(actions.value, s_actions, rtol=1E-5)
            assert np.allclose(angles.value, s_angles, rtol=1E-5)
            assert np.allclose(freqs.value, s_freqs, rtol=1E-5)

            # logger.debug("Plotting orbit...")
            # fig = plot_orbits(w, marker='.', alpha=0.2, linestyle='none')
            # fig.savefig(str(self.plot_path.join("orbit_{}.png".format(n))))

            # fig = plot_angles(t,angles,freqs)
            # fig.savefig(str(self.plot_path.join("angles_{}.png".format(n))))

            # fig = plot_angles(t,s_angles,s_freqs)
            # fig.savefig(str(self.plot_path.join("angles_sanders_{}.png".format(n))))

            # plt.close('all')

            # print("Plots saved at:", self.plot_path)

class TestActions(ActionsBase):

    def setup(self):
        self.plot_path = self.tmpdir.mkdir("normal")

        self.units = galactic
        self.potential = LeeSutoTriaxialNFWPotential(v_c=0.2, r_s=20.,
                                                     a=1., b=0.77, c=0.55,
                                                     units=galactic)
        self.N = 8
        np.random.seed(42)
        w0 = isotropic_w0(N=self.N)
        n_steps = 20000

        # integrate orbits
        orbit = self.potential.integrate_orbit(w0, dt=2., n_steps=n_steps,
                                               Integrator=DOPRI853Integrator)
        self.orbit = orbit
        self.t = orbit.t.value
        self.w = orbit.w()

# TODO: need to fix this -- or assess whether needed?
# class TestHardActions(ActionsBase):

#     def setup(self):
#         self.plot_path = self.tmpdir.mkdir("hard")

#         self.units = (u.kpc, u.Msun, u.Myr)
#         params = {'disk': {'a': 6.5, 'b': 0.26, 'm': 65000000000.0},
#                   'bulge': {'c': 0.3, 'm': 20000000000.0},
#                   'halo': {'psi': 1.570796, 'theta': 1.570796, 'phi': 1.570796,
#                            'a': 1., 'b': 0.77, 'c': 0.61, 'r_s': 30.0, 'v_c': 0.22}}
#         self.potential = PW14Potential(**params)
#         self.N = 25

#         w0 = np.loadtxt(os.path.join(test_data_path, "w0.txt"))
#         n_steps = 200000

#         if not os.path.exists(os.path.join(test_data_path, "w_hard.npy")):
#             logger.debug("Integrating orbits")
#             t,w = self.potential.integrate_orbit(w0, dt=0.2, n_steps=n_steps, Integrator=DOPRI853Integrator)

#             logger.debug("Saving orbits")
#             np.save(os.path.join(test_data_path, "t_hard.npy"), t)
#             np.save(os.path.join(test_data_path, "w_hard.npy"), w)
#         else:
#             logger.debug("Loaded orbits")
#             t = np.load(os.path.join(test_data_path, "t_hard.npy"))
#             w = np.load(os.path.join(test_data_path, "w_hard.npy"))

#         self.t = t
#         self.w = w

def test_compare_action_prepare():

    from ..actionangle import _action_prepare, _angle_prepare

    logger.setLevel(logging.ERROR)
    AA = np.random.uniform(0., 100., size=(1000,6))
    t = np.linspace(0., 100., 1000)

    act_san,n_vectors = solver.solver(AA, N_max=6, symNx=2)
    A2,b2,n = _action_prepare(AA.T, N_max=6, dx=2, dy=2, dz=2)
    act_apw = np.array(solve(A2,b2))

    ang_san = solver.angle_solver(AA, t, N_max=6, symNx=2, sign=1)
    A2,b2,n = _angle_prepare(AA.T, t, N_max=6, dx=2, dy=2, dz=2)
    ang_apw = np.array(solve(A2,b2))

    assert np.allclose(act_apw, act_san)
    # assert np.allclose(ang_apw, ang_san)

    # TODO: this could be critical -- why don't our angles agree?



# coding: utf-8

""" 5th order Runge-Kutta integration. """

from __future__ import division, print_function


# Third-party
import numpy as np

# Project
from ..core import Integrator
from ..timespec import parse_time_specification

__all__ = ["RK5Integrator"]

# These are the Dormand-Prince parameters for embedded Runge-Kutta methods
A = np.array([0.0, 0.2, 0.3, 0.6, 1.0, 0.875])
B = np.array([[        0.0,        0.0,         0.0,            0.0,        0.0],
            [      1./5.,        0.0,         0.0,            0.0,        0.0],
            [      3./40.,    9./40.,         0.0,            0.0,        0.0],
            [      3./10.,   -9./10.,       6./5.,            0.0,        0.0],
            [    -11./54.,     5./2.,    -70./27.,        35./27.,        0.0],
            [1631./55296., 175./512., 575./13824., 44275./110592., 253./4096.]])
C = np.array([37./378., 0., 250./621., 125./594., 0., 512./1771.])
D = np.array([2825./27648., 0., 18575./48384., 13525./55296., 277./14336., 1./4.])

class RK5Integrator(Integrator):
    r"""
    Initialize a 5th order Runge-Kutta integrator given a function for
    computing derivatives with respect to the independent variables. The
    function should, at minimum, take the independent variable as the
    first argument, and the coordinates as a single vector as the second
    argument. For notation and variable names, we assume this independent
    variable is time, t, and the coordinate vector is named x, though it
    could contain a mixture of coordinates and momenta for solving
    Hamilton's equations, for example.

    .. seealso::

        - http://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

    Parameters
    ----------
    func : func
        A callable object that computes the phase-space coordinate
        derivatives with respect to the independent variable at a point
        in phase space.
    func_args : tuple (optional)
        Any extra arguments for the function.
    func_units : `~gala.units.UnitSystem` (optional)
        If using units, this is the unit system assumed by the
        integrand function.

    """

    def step(self, t, w, dt):
        """ Step forward the vector w by the given timestep.

            Parameters
            ----------
            dt : numeric
                The timestep to move forward.
        """

        # Runge-Kutta Fehlberg formulas (see: Numerical Recipes)
        F = lambda t,w: self.F(t,w,*self._func_args)

        n = len(w)
        K = np.zeros((6,)+w.shape)
        K[0] = dt * F(t, w)
        K[1] = dt * F(t + A[1]*dt, w + B[1][0]*K[0])
        K[2] = dt * F(t + A[2]*dt, w + B[2][0]*K[0] + B[2][1]*K[1])
        K[3] = dt * F(t + A[3]*dt, w + B[3][0]*K[0] + B[3][1]*K[1] + B[3][2]*K[2])
        K[4] = dt * F(t + A[4]*dt, w + B[4][0]*K[0] + B[4][1]*K[1] + B[4][2]*K[2] + B[4][3]*K[3])
        K[5] = dt * F(t + A[5]*dt, w + B[5][0]*K[0] + B[5][1]*K[1] + B[5][2]*K[2] + B[5][3]*K[3] + B[5][4]*K[4])

        # shift
        dw = np.zeros_like(w)
        for i in range(6):
            dw = dw + C[i]*K[i]

        return w + dw

    def run(self, w0, mmap=None, **time_spec):

        # generate the array of times
        times = parse_time_specification(self._func_units, **time_spec)
        n_steps = len(times)-1
        dt = times[1]-times[0]

        w0_obj, w0, ws = self._prepare_ws(w0, mmap, n_steps=n_steps)

        # Set first step to the initial conditions
        ws[:,0] = w0
        w = w0.copy()
        for ii in range(1,n_steps+1):
            w = self.step(times[ii], w, dt)
            ws[:,ii] = w

        return self._handle_output(w0_obj, times, ws)
